# Copyright 2014 Arndt Droullier, Nive GmbH. All rights reserved.
# Released under GPL3. See license.txt
#

from nive.definitions import ModuleConf, Conf


class FilenameLookup(object):
    """
    Enables readable url path names instead of ids for object traversal.
    Names are stored as meta.pool_filename.

    Extensions like *.html* are not stored. Path matching works independent 
    from extensions.
    """
    def Init(self):
        self.__name__ = self.meta.pool_filename or str(self.id)

    def __getitem__(self, id):
        """
        Traversal lookup based on object.pool_filename and object.id. Trailing extensions 
        are ignored.
        `file` is a reserved name and used in the current object to map file downloads. 
        """
        if id == "file":
            raise KeyError(id)
        id = id.split(".")
        if len(id)>2:
            id = (".").join(id[:-1])
        else:
            id = id[0]
        try:
            id = int(id)
        except:
            name = id
            id = 0
            if name:
                id = self.root.search.FilenameToID(name, self.id)
            if not id:
                raise KeyError(id)
        o = self.GetObj(id)
        if not o:
            raise KeyError(id)
        return o



def SetupFilenameLookup(app, pyramidConfig):
    # get all roots and add extension
    extension = "nive.extensions.filename.FilenameLookup"
    def add(confs):
        for c in confs:
            e = c.extensions
            if e and extension in e:
                continue
            if e == None:
                e = []
            if isinstance(e, tuple):
                e = list(e)
            e.append(extension)
            c.unlock()
            c.extensions = tuple(e)
            c.lock()

    add(app.configurationQuery.GetAllRootConfs())
    add(app.configurationQuery.GetAllObjectConfs())


configuration = ModuleConf(
    id = "filenameLookup",
    name = " Enables readable url path names instead of ids for object traversal",
    context = "nive.extensions.filename",
    events = (Conf(event="startRegistration", callback=SetupFilenameLookup),),
)



# Copyright 2012, 2013 Arndt Droullier, Nive GmbH. All rights reserved.
# Released under GPL3. See license.txt
#
__doc__ = """
This file contains the main application functionality. The application handles configuration
registration, roots and the database. Also it provides convenient functions to lookup all
different kinds of configurations. 

Application.modules stores configuration objects registered by calling `Register()`. 

See Application.Startup() for the main entry point and connected events. 

Not to be mixed up with pyramid applications or projects. It is possible to use multiple nive
applications in a single pyramid app.
"""

import logging
import pytz
import uuid

from time import time

from zope.interface.registry import Components

from nive.utils.dataPool2.structure import PoolStructure

from nive.definitions import MetaTbl
from nive.definitions import IAppConf
from nive.definitions import ConfigurationError

from nive.registration import Registration, AppConfigurationQuery
from nive.events import Events
from nive.helper import ResolveConfiguration, GetClassRef, ClassFactory
from nive.utils.utils import SortConfigurationList
from nive.tool import _IGlobal, _GlobalObject
from nive.search import Search


from nive.definitions import implementer
from nive.definitions import IApplication



@implementer(IApplication)
class Application(Events):
    """
    nive Application implementaion.

    SQLite example:
    ::

        appConf = AppConf(id="website",
                          title="nive app")
        dbConf = DatabaseConf(context="Sqlite3",
                              fileRoot="data",
                              dbName="data/ewp.db")
        appConf.modules.append(dbConf)

    MySql example:
    ::

        appConf = AppConf(id="website",
                          title="nive app")
        dbConf = DatabaseConf(context="MySql",
                              fileRoot="data",
                              dbName="nive",
                              host="localhost",
                              user="user",
                              password="password")
        appConf.modules.append(dbConf)

    Requires (Configuration, AppFactory, Events.Events)

    The startup process is handled by nive.portal.portal on application startup.
    During startup the system fires four events in the following order:

    - startup(app)
    - startRegistration(app, pyramidConfig)
    - finishRegistration(app, pyramidConfig)
    - run(app)

    All acl definitions are mapped to self.__acl__.

    ViewModuleConf.acl values are added first, AppConf.acl are appended to the end of the list.
    This might make it hard to replace single permissions defined by view modules because pyramid
    uses the first matching definition in the list.

    To disable any view module acl definitions or replace single permissions use the `FinishRegistration`
    event and post process acls as needed.
    """

    def __init__(self, configuration=None):
        """
        Use configuration and dbConfiguration objects to setup your application.

        Events:
        - init(configuration)
        """
        self.registry = Components()
        self.configuration = configuration
        self.dbConfiguration = None
        self.pytimezone = None
        # set id for logging
        if configuration:
            self.id = configuration.id
        else:
            self.id = __name__
        self.uid = str(uuid.uuid4())

        self.__name__ = ""
        self.__parent__ = None
        self.__acl__ = []

        # development
        self.debug = False
        self.reloadExtensions = False

        # default root name (root.configuration.id)
        self._defaultRoot = ""
        # cache database structure
        self._structure = PoolStructure()
        self._dbpool = None

        self.log = logging.getLogger(self.id)
        self.log.debug("Initialize %s", repr(configuration))
        self.Signal("init", configuration=configuration)

        self.starttime = 0

    def __del__(self):
        self.Close()

    def __getitem__(self, name):
        """
        This function is called by url traversal and looks up root objects for the corresponding
        name.
        """
        name = name.split(".")[0]
        o = self.GetRoot(name)
        if o is not None and o.configuration.urlTraversal:
            return o
        raise KeyError(name)

    # Properties -----------------------------------------------------------

    @property
    def root(self):
        """
        Get default root object. Use `GetRoot(rootname)` if multiple roots used.

        Events:
        - loadRoot(root)
        - loadFromCache() called for the root

        returns root object
        """
        return self.factory.GetRootObj(self._defaultRoot)

    @property
    def db(self):
        """ returns datapool object    """
        return self._dbpool

    @property
    def portal(self):
        """ returns the portal """
        return self.__parent__

    @property
    def factory(self):
        """ returns the obj factory instance. """
        return AppFactory(self)

    @property
    def configurationQuery(self):
        """ returns the app configuration/registry query instance """
        return AppConfigurationQuery(self)

    @property
    def app(self):
        """ returns itself. for compatibility. """
        return self

    @property
    def search(self):
        return Search(self.root)

    @property
    def rootname(self):
        """ returns the name of the default root. """
        return self._defaultRoot

    def obj(self, id, rootname="", **kw):
        """ returns object    """
        root = self.factory.GetRootObj(rootname)
        if root is None:
            return None
        return root.obj(id, **kw)


    # Events and Setup ------------------------------------------------------------------

    def Startup(self, pyramidConfig, debug=False, cachedDbConnection=None):
        self.Setup(debug=debug)
        self.StartRegistration(pyramidConfig)
        self.FinishRegistration(pyramidConfig, cachedDbConnection=cachedDbConnection)
        self.Run()

    def Setup(self, debug=False):
        """
        Called by nive.portal.portal on application startup.
        Sets up the registry.

        Events:

        - startup(app)
        """
        self.starttime = time()
        self.log.debug("Startup with debug=%s", str(debug))
        self.debug = debug
        self.Signal("startup", app=self)
        self.SetupApplication()

    def SetupApplication(self):
        """
        Loads self.configuration, includes modules and updates meta fields.
        """
        if not self.configuration:
            raise ConfigurationError("Configuration is empty")
        c = self.configuration
        # special values
        if c.get("id") and not self.__name__:
            self.__name__ = c.id
        if c.get("acl"):
            self.__acl__ = tuple(c.acl)
        if c.get("dbConfiguration"):
            self.dbConfiguration = c.dbConfiguration
        tz = self.configuration.get("timezone")
        if tz is not None:
            self.pytimezone = pytz.timezone(tz)

    def StartRegistration(self, pyramidConfig):
        """
        Processes all registered components and commits view definitions to
        the pyramid application configuration.

        Events:

        - startRegistration(app, pyramidConfig)
        """
        # register modules `configuration.modules`
        reg = Registration(self)
        reg.RegisterComponents(self.configuration)
        # signal before committing views
        self.Signal("startRegistration", app=self, pyramidConfig=pyramidConfig)
        # register pyramid views and translations
        if pyramidConfig:
            reg.RegisterViewModules(pyramidConfig)
            pyramidConfig.commit()
            reg.RegisterViews(pyramidConfig)
            pyramidConfig.commit()
            reg.RegisterTranslations(pyramidConfig)
            pyramidConfig.commit()

        # register groups
        portal = self.portal
        if portal and hasattr(portal, "RegisterGroups"):
            portal.RegisterGroups(self)

    def FinishRegistration(self, pyramidConfig, cachedDbConnection=None):
        """
        Finishes the registration process and cashes the database structure.
        Configurations will have a write lock after the registration is finished.

        `cachedDbConnection` is a nive.utils.dataPool2.connection instance to
        avoid reconnects on application startup. The connection can be accessed
        during runtime through ``self.db.usedconnection``.

        Events:

        - finishRegistration(app, pyramidConfig)
        """
        self.Signal("finishRegistration", app=self, pyramidConfig=pyramidConfig)
        self.log.debug('Finished registration.')

        # reload database structure
        self._LoadStructure(forceReload=True)
        self._dbpool = self.factory.GetDataPoolObj(cachedDbConnection)

        # test and create database fields
        if self.debug:
            self.GetTool("nive.tools.dbStructureUpdater", self).Run()
            result, report = self.TestDB()
            if result:
                self.log.info('Database test result: %s %s', str(result), report)
            else:
                self.log.error('Database test failure: (%s) %s', str(result), report)

        self._Lock()

    def Run(self):
        """
        Signals the application is up and running.

        Events:

        - run(app)
        """
        # start
        self.Signal("run", app=self)
        self.log.info('Application %s running. Startup time: %.05f.', self.__class__.__name__, time() - self.starttime)
        self.id = self.configuration.id
        self.log = logging.getLogger(self.id)

    def Close(self):
        """
        Close database and roots.

        Events:

        - close()
        """
        self.Signal("close")
        self.factory.CloseRootObj()
        if self._dbpool:
            self._dbpool.Close()

    def Shutdown(self):
        """
        Free registry and cached database structure.
        """
        self.registry = None
        self._structure = None
        self.configuration = None
        self.dbConfiguration = None
        self.__parent__ = None

    # Content and root objects -----------------------------------------------------------

    def GetRoot(self, name=""):
        """
        Returns the data root object. If name is empty the default root is returned.

        Events:
        - loadRoot(root)
        - loadFromCache() called for the root

        returns root object
        """
        return self.factory.GetRootObj(name)

    def GetRoots(self):
        """
        Returns all root objects.

        returns list
        """
        return [self.GetRoot(r.id) for r in self.configurationQuery.GetAllRootConfs()]

    def GetRootNames(self):
        """
        Get all root object names (root.configuration.id).

        returns list
        """
        return [r["id"] for r in self.configurationQuery.GetAllRootConfs()]

    def GetTool(self, toolID, contextObject=None):
        """
        Load tool object. *toolID* must be the tool.id or dotted python name.

        returns the tool object.
        """
        return self.factory.GetToolObj(toolID, contextObject or _GlobalObject())

    def GetWorkflow(self, wfProcID, contextObject=None):
        """
        Load workflow process. *wfProcID* must be the wf.id or dotted python name.

        returns the workflow object
        """
        if not self.configuration.workflowEnabled:
            return None
        return self.factory.GetWfObj(wfProcID, contextObject or _GlobalObject())

    # Groups -------------------------------------------------------------

    def GetGroups(self, sort="name", visibleOnly=False):
        """
        Get a list of configured groups.

        returns list
        """
        configuration = self.configuration
        if not configuration.groups:
            return []
        if not visibleOnly:
            return SortConfigurationList(configuration.groups, sort)
        c = [a for a in configuration.groups if not a.get("hidden")]
        return SortConfigurationList(c, sort)

    def GetGroupName(self, groupID):
        """
        Get group name for id.

        returns string
        """
        g = [d for d in self.app.configuration.groups if d["id"] == groupID]
        if not g:
            return ""
        return g[0]["name"]

    # Data Pool and database connections -------------------------------------------------------------------

    def GetDB(self):
        """
        Create data pool object and database connection.

        returns the datapool object
        """
        return self._dbpool

    def Query(self, sql, values=()):
        """
        Start a sql query on the database.

        returns tuple or none
        """
        db = self.db
        r = db.Query(sql, values)
        return r

    def TestDB(self):
        """
        Test database connection for errors. Returns state and error message.

        returns bool, message
        """
        try:
            db = self.db
            if not db.connection.IsConnected():
                db.connection.connect()
                if not db.connection.IsConnected():
                    return False, "No connection"
            #self.Query("select id from pool_meta where id =1")
            return True, "OK"

        except Exception as err:
            return False, str(err)

    def NewConnection(self):
        """
        Creates a new database connection. Works before startup and application setup.
        """
        return self.factory.GetConnection()

    def NewDBApi(self):
        """
        Creates a raw database connection (dbapi). Works before startup and application setup.
        """
        return self.factory.GetDBApi()

    def NewModule(self, queryFor, name, context=None):
        """
        Query for configuration and lookup class reference. Does not call __init__ for
        the new class.

        returns class or None
        """
        c = self.configurationQuery.QueryConfByName(queryFor, name, context=context)
        if c is None:
            return None
        return ClassFactory(c)



    # System value storage -------------------------------------------------------------

    def StoreSysValue(self, key, value):
        """
        Stores a value in `pool_sys` table. Value must be a string of any size.
        """
        db = self.db
        db.UpdateFields("pool_sys", key, {"id": key, "value":value, "ts":time()}, autoinsert=True)
        db.Commit()

    def LoadSysValue(self, key):
        """
        Loads the value stored as `key` from `pool_sys` table.
        """
        db = self.db
        sql, values = db.FmtSQLSelect(["value", "ts"], parameter={"id":key}, dataTable="pool_sys", singleTable=1)
        r = db.Query(sql, values)
        if not r:
            return None
        return r[0][0]

    def DeleteSysValue(self, key):
        """
        Deletes a single system value
        """
        db = self.db
        db.DeleteRecords("pool_sys", {"id": key})
        db.Commit()


    def _Lock(self):
        # lock all configurations
        # adapters
        for a in self.app.registry.registeredAdapters():
            try:
                a.factory.lock()
            except:
                pass
        for a in self.app.registry.registeredUtilities():
            try:
                a.component.lock()
            except:
                pass


    def _LoadStructure(self, forceReload=False):
        """
        returns dictionary containing database tables and columns
        """
        app = self.app
        if forceReload:
            app._structure = PoolStructure()
        if not app._structure.IsEmpty():
            return app._structure

        structure = {}
        fieldtypes = {}
        meta = app.configurationQuery.GetAllMetaFlds(ignoreSystem=False)
        m = []
        f = {}
        for fld in meta:
            f[fld.id] = fld.datatype
            if fld.datatype == "file":
                continue
            m.append(fld.id)
        structure[MetaTbl] = m
        fieldtypes[MetaTbl] = f

        types = app.configurationQuery.GetAllObjectConfs()
        for ty in types:
            t = []
            f = {}
            for fld in app.configurationQuery.GetAllObjectFlds(ty.id):
                f[fld.id] = fld.datatype
                if fld.datatype == "file":
                    continue
                t.append(fld.id)
            structure[ty.dbparam] = t
            fieldtypes[ty.dbparam] = f

        app._structure.Init(structure, fieldtypes, m, app.configuration.frontendCodepage, self.pytimezone)
        # reset cached db
        if app._dbpool is not None:
            app._dbpool.Close()
        return structure


    # bw [3] tests
    def Register(self, conf, **kw):
        iface, conf = ResolveConfiguration(conf)
        if iface == IAppConf:
            self.configuration = conf
        Registration(self).RegisterComponents(conf, **kw)


class AppFactory(object):
    """
    Internal class for dynamic object creation and caching.

    Requires:
    - Application
    """

    def __init__(self, app):
        self.app = app

    def GetDataPoolObj(self, cachedDbConnection=None):
        """
        creates the database object
        """
        app = self.app
        if not app.dbConfiguration:
            raise ConfigurationError("Database configuration empty. application.dbConfiguration is None.")
        poolTag = app.dbConfiguration.context
        if not poolTag:
            raise TypeError("Database type not set. application.dbConfiguration.context is empty. Use Sqlite or Mysql!")
        elif poolTag.lower() in ("sqlite", "sqlite3"):
            poolTag = "nive.utils.dataPool2.sqlite.sqlite3Pool.Sqlite3"
        elif poolTag.lower() == "mysql":
            poolTag = "nive.utils.dataPool2.mysql.mySqlPool.MySql"
        elif poolTag.lower() in ("postgres", "postgresql"):
            poolTag = "nive.utils.dataPool2.postgres.postgreSqlPool.PostgreSql"

        # if a database connection other than the default is configured
        if cachedDbConnection is not None:
            connObj = cachedDbConnection
        elif app.dbConfiguration.connection:
            connObj = GetClassRef(app.dbConfiguration.connection, app.reloadExtensions, True, None)
            connObj = connObj(config=app.dbConfiguration, connectNow=False)
        else:
            connObj = None

        dbObj = GetClassRef(poolTag, app.reloadExtensions, True, None)
        conn = app.dbConfiguration
        dbObj = dbObj(connection=connObj,
                      connParam=conn,  # use the default connection defined in db if connection is none
                      structure=app._structure,
                      root=conn.fileRoot,
                      useTrashcan=conn.useTrashcan,
                      dbCodePage=conn.dbCodePage,
                      timezone=app.pytimezone,
                      debug=conn.querylog[0],
                      log=conn.querylog[1])
        return dbObj

    def GetConnection(self):
        """
        Creates a new database connection. Works before startup and application setup.
        """
        app = self.app
        if app._dbpool is not None:
            try:
                return app._dbpool.usedconnection
            except AttributeError:
                pass

        cTag = app.dbConfiguration.connection
        poolTag = app.dbConfiguration.context
        if cTag:
            connObj = GetClassRef(cTag, app.reloadExtensions, True, None)
            connObj = connObj(config=app.dbConfiguration, connectNow=False)
            return connObj
        if app._dbpool is not None:
            return app._dbpool.CreateConnection(app.dbConfiguration)
        if not cTag and not poolTag:
            raise TypeError(
                "Database connection type not set. application.dbConfiguration.connection and application.dbConfiguration.context is empty. Use Sqlite or Mysql!")
        poolTag = app.dbConfiguration.context
        if not poolTag:
            raise TypeError("Database type not set. application.dbConfiguration.context is empty. Use Sqlite or Mysql!")
        elif poolTag.lower() in ("sqlite", "sqlite3"):
            poolTag = "nive.utils.dataPool2.sqlite.sqlite3Pool.Sqlite3"
        elif poolTag.lower() == "mysql":
            poolTag = "nive.utils.dataPool2.mysql.mySqlPool.MySql"
        elif poolTag.lower() in ("postgres", "postgresql"):
            poolTag = "nive.utils.dataPool2.postgres.postgreSqlPool.PostgreSql"
        dbObj = GetClassRef(poolTag, app.reloadExtensions, True, None)
        return dbObj._DefaultConnection(config=app.dbConfiguration, connectNow=False)

    def GetDBApi(self):
        """
        Creates a new and raw database connection. Works before startup and application setup.
        """
        conn = self.GetConnection()
        if conn:
            return conn.PrivateConnection()
        return None

    def GetRootObj(self, name):
        """
        creates the root object
        """
        app = self.app
        if not name:
            name = app.rootname
        useCache = app.configuration.useCache
        if isinstance(name, str):
            cachename = "_c_root" + name
            if useCache and hasattr(app, cachename) and getattr(app, cachename) is not None:
                rootObj = getattr(app, cachename)
                rootObj.Signal("loadFromCache")
                return rootObj
            rootConf = app.configurationQuery.GetRootConf(name)
        else:
            rootConf = name

        if rootConf is None:
            return None

        name = rootConf.id
        cachename = "_c_root" + name
        if useCache and hasattr(app, cachename) and getattr(app, cachename) is not None:
            rootObj = getattr(app, cachename)
            rootObj.Signal("loadFromCache")
        else:
            rootObj = ClassFactory(rootConf, app.reloadExtensions, True, base=None)
            rootObj = rootObj(name, app, rootConf)
            if rootObj and useCache:
                setattr(app, cachename, rootObj)
        app.Signal("loadRoot", rootObj)
        return rootObj

    def CloseRootObj(self, name=None):
        """
        close root objects. if name = none, all are closed.
        """
        app = self.app
        if not name:
            if app.registry is None:
                return
            n = app.GetRootNames()
        else:
            n = (name,)
        for name in n:
            cachename = "_c_root" + name
            try:
                o = getattr(app, cachename)
                o.Close()
                setattr(app, cachename, None)
            except AttributeError:
                pass

    def GetToolObj(self, name, contextObject):
        """
        creates the tool object
        """
        app = self.app
        if isinstance(name, str):
            conf = app.configurationQuery.GetToolConf(name, contextObject)
            if isinstance(conf, (list, tuple)):
                conf = conf[0]
        else:
            conf = name
        if not conf:
            iface, conf = ResolveConfiguration(name)
            if not conf:
                raise ImportError("Tool not found. Please load the tool by referencing the tool id. (%s)" % (str(name)))
        tag = conf.context
        toolObj = GetClassRef(tag, app.reloadExtensions, True, None)
        toolObj = toolObj(conf, app)
        if not _IGlobal.providedBy(contextObject):
            toolObj.__parent__ = contextObject
        return toolObj

    def GetWfObj(self, name, contextObject):
        """
        creates the root object
        """
        app = self.app
        wfConf = None
        if isinstance(name, str):
            wfConf = app.configurationQuery.GetWorkflowConf(name, contextObject)
            if isinstance(wfConf, (list, tuple)):
                wfConf = wfConf[0]
        if not wfConf:
            iface, wfConf = ResolveConfiguration(name)
            if not wfConf:
                raise ImportError(
                    "Workflow process not found. Please load the workflow by referencing the process id. (%s)" % (
                        str(name)))

        wfTag = wfConf.context
        wfObj = GetClassRef(wfTag, app.reloadExtensions, True, None)
        wfObj = wfObj(wfConf, app)
        if not _IGlobal.providedBy(contextObject):
            wfObj.__parent__ = contextObject
        return wfObj




# -*- coding: utf-8 -*-

import unittest

from nive.helper import FormatConfTestFailure

from nive.components.iface import view



class TestConf(unittest.TestCase):

	def test_conf1(self):
		r=view.configuration.test()
		if not r:
			return
		self.fail(FormatConfTestFailure(r))






# Copyright 2012, 2013 Arndt Droullier, Nive GmbH. All rights reserved.
# Released under GPL3. See license.txt
#

__doc__ = """
Basic view class for nive object views. 

Use this class as base for custom view definitions for file download support,
data rendering, url generation, http headers and user lookup. 
"""


import os
import time
import collections
try:
    import simplejson as json
except:
    import json
from datetime import datetime
from email.utils import formatdate

from pyramid.response import Response
from pyramid.renderers import render_to_response, get_renderer, render
from pyramid.url import static_url, resource_url
from pyramid.view import render_view
from pyramid.security import authenticated_userid
from pyramid.i18n import get_localizer

from pyramid.httpexceptions import HTTPNotFound, HTTPFound, HTTPOk, HTTPForbidden, HTTPException
from pyramid.exceptions import NotFound

from nive.i18n import _
from nive import helper
from nive.utils.utils import ConvertToStr, ConvertListToStr, ConvertToDateTime
from nive.utils.utils import FmtSeconds, FormatBytesForDisplay, CutText, GetMimeTypeExtension
from nive import FileNotFound
from nive.definitions import ViewConf
from nive.definitions import IViewModuleConf
from nive.definitions import ConfigurationError



class Unauthorized(Exception):
    """
    Raised by failed logins or insufficient permissions
    """
    


class BaseView(object):
    """    """
    __configuration__ = None

    
    def __init__(self, context, request):
        self.context = context
        self.request = request
        self.appRequestKeys = []  # todo ?
        self.fileExpires = 3600
        self._t = time.time()

    @property
    def configuration(self):
        """
        View module configuration

        If views are registered by using `ViewModuleConf.views` the ViewModuleConf is automatically
        available as class attribute of the view instance. Any specific configuration values are
        avalilable directly through the configuration. E.g.

            index = self.viewModule.template

        returns ViewModuleConf
        """
        if self.__configuration__ is not None:
            return self.__configuration__()
        return None

    # url handling ----------------------------------------------------------------

    def Url(self, resource=None):
        """
        Generates the default url for the resource. Includes the file extension if one specified 
        for the object.
        
        If resource is None the current context object is used as resource.

        returns url
        """
        resource=resource or self.context
        url = resource_url(resource, self.request)
        if not hasattr(resource, "extension") or not resource.extension:
            return url
        # add extension if it is not the virtual root
        if self.request.virtual_root == resource:
            return url
        return "%s.%s" % (url[:-1], resource.extension)

    def FolderUrl(self, resource=None):
        """
        Generates the default url for the resource without extension and trailing '/'. 
        
        If resource is None the current context object is used as resource.

        returns url
        """
        return resource_url(resource or self.context, self.request)

    def StaticUrl(self, file):
        """
        Generates the url for a static file.
        *file* is the filename to look the url up for. E.g. for the default static directory use ::
        
            <link tal:attributes="href view.StaticUrl('layout2.css')" 
                  rel="stylesheet" type="text/css" media="all" >
                  
         to reference a file in a different directory you must include the python module like ::

            <link tal:attributes="href view.StaticUrl('my_app.design:static/layout2.css')" 
                  rel="stylesheet" type="text/css" media="all" >
                  
         absolute http urls can also be used based on the url prefix http, https and / ::

            <link tal:attributes="href view.StaticUrl('/assets/layout2.css')" 
                  rel="stylesheet" type="text/css" media="all" >
            <link tal:attributes="href view.StaticUrl('http://myserver.com/assets/layout2.css')" 
                  rel="stylesheet" type="text/css" media="all" >
            <link tal:attributes="href view.StaticUrl('https://myserver.com/assets/layout2.css')" 
                  rel="stylesheet" type="text/css" media="all" >
                  
        returns url
        """
        if file is None:
            return ""
        if file.startswith(("http://","https://","/")):
            return file
        if not ":" in file:
            conf = self.configuration
            if conf and conf.static:
                if conf.static.endswith(("/",":")):
                    file = conf.static + file
                else:
                    file = "%s/%s" % (conf.static, file)
                if file.startswith(("http://","https://","/")):
                    return file
        return static_url(file, self.request)

    def FileUrl(self, fieldID, resource=None):
        """
        Generates the file url for the file contained in resource. The resource must have the 'file'
        view included (IFile). If the url is called the download is mapped to ``View.file``. 
        
        If resource is None the current context object is used as resource.

        returns url
        """
        if not resource:
            resource=self.context
        file = resource.files.get(fieldID)
        if not file:
            return ""
        return "%sfile/%s" % (self.Url(resource), file.filename)

    def PageUrl(self, resource=None, usePageLink=0, addAnchor=True):
        """
        Generates the default page url for the resource with extension. If resource is a page element
        the page containing this element is used for the url. 
        
        If `resource` is None the current context object is used as resource.

        If `addAnchor` is true and resource is not the actula page the url is generated for, the function 
        will append a anchor for the resource to the url based on the id.
        
        returns url
        """
        resource = resource or self.context
        try:
            page = resource.GetPage()
        except AttributeError:
            page = self.context
        link = page.data.get("pagelink")
        if usePageLink and link:
            return self.ResolveUrl(link, resource)

        url = resource_url(page, self.request)
        if hasattr(page, "extension") and page.extension:
            # add extension if it is not the virtual root
            if self.request.virtual_root != page:
                url = "%s.%s" % (url[:-1], page.extension)

        if not addAnchor or resource == page:
            return url
        # add anchor if resource is a element
        return "%s#nive-element%d"%(url, resource.id)

    def CurrentUrl(self, retainUrlParams=False, preserveHost=1):
        """
        Returns the current url that triggered this request. Url parameter are removed by
        default.
        
        returns url
        """
        if retainUrlParams:
            url = self.request.url
        else:
            url = self.request.url.split("?")[0]
        if preserveHost:
            return url
        return "/"+"/".join(url.split("/")[3:])


    def ResolveUrl(self, url, context=None):
        """
        Resolve a string to url for object or the current context.

        Url placeholders to be used as url

        - obj_url
        - obj_folder_url
        - root_url
        - root_folder_url
        - app_url
        - app_folder_url
        - parent_url
        - page_url
        - page_url_anchor

        The url can also be a callable. It is called with two parameters
        `context` and `view class instance`. E.g. ::

            def makeUrl(context, view):
                return view.GetUrl(context) + "?query=" + view.GetFormValue('query')

        The placeholder can be used as the root and be followed by a path e.g. ``obj_folder_url/edit``.
        The remaining path after the slash will be preserved.

        returns url
        """
        if url is None:
            return ""
        if not context:
            context = self.context

        # todo [3]
        # DeprecationWarning: Using or importing the ABCs from 'collections' instead of from 'collections.abc' is deprecated, and in 3.8 it will stop working
        #   if isinstance(url, collections.Callable):
        if isinstance(url, collections.abc.Callable):
            return url(context, self)

        parts = url.split("/")
        url = parts[0]
        if url == "page_url":
            url = self.PageUrl(context)
        elif url == "page_url_anchor":
            url = self.PageUrl(context, addAnchor=True)
        elif url == "obj_url":
            url = self.Url(context)
        elif url == "obj_folder_url":
            url = self.FolderUrl(context)
        elif url == "parent_url":
            url = self.Url(context.parent)
        elif url == "root_url":
            url = self.Url(context.root)
        elif url == "root_folder_url":
            url = self.FolderUrl(context.root)
        elif url == "app_url":
            url = self.Url(context.app)
        elif url == "app_folder_url":
            url = self.FolderUrl(context.app)
        if len(parts)>1:
            parts[0] = url[:-1]
            return "/".join(parts)
        return url
    
    
    def ResolveLink(self, link):
        """
        Resolve the link to a valid (page) URL. If the link is an existing object.id the url to the 
        page containing this object is returned.
        
        returns url
        """
        try:
            i = int(link)
            o = self.context.root.obj(i)
            if not o:
                return link
            try:
                return self.PageUrl(o)
            except:
                return self.Url(o)
        except:
            return link

    
    # Response and headers (defined on module level below -------------------------------------------

    def SendResponse(self, data, mime="text/html", filename=None, raiseException=True, status=None, headers=None):
        """
        Creates a response with data as body. If ``raiseException`` is true the function
        will raise a HTTPOk Exception with data as body. A custom response status can only be passed
        if `raiseException` is False.
        
        If filename is not none the response will extended with a ``attachment; filename=filename``
        header.
        """
        return SendResponse(data, mime=mime, filename=filename, raiseException=raiseException, status=status, headers=headers)
        
        
    def Redirect(self, url, messages=None, slot="", raiseException=True, refresh=True, force=False):
        """
        Redirect to the given URL. Messages are stored in session and can be accessed
        by calling ``request.session.pop_flash()``. Messages are added by calling
        ``request.session.flash(m, slot)``.

        `Redirect` automatically detects ajax requests and calls `Relocate` in such a case to prevent a page refresh.
        Set `force=True` to force redirects even for ajax requests.
        """
        return Redirect(url, self.request, messages=messages, slot=slot, raiseException=raiseException, refresh=refresh, force=force)
    

    def Relocate(self, url, messages=None, slot="", raiseException=True, refresh=True):
        """
        Returns messages and X-Relocate header in response.
        If raiseException is True HTTPOk is raised with empty body.
        """
        return Relocate(url, self.request, messages=messages, slot=slot, raiseException=raiseException, refresh=refresh)


    def ResetFlashMessages(self, slot=""):
        """
        Removes all messages stored in session.
        """
        ResetFlashMessages(self.request, slot)
        
    
    def AddHeader(self, name, value, append=False):
        """
        Add a additional response header value.
        """
        AddHeader(self.request, name, value, append)

    
    # render other elements and objects ---------------------------------------------
    
    def index_tmpl(self, path=None):
        #if path:
        #    return get_renderer(path).implementation()
        if not path:
            conf = self.configuration
            if not conf or not conf.template:
                return None
            path = conf.template
        tmpl = self._LookupTemplate(path)
        return get_renderer(tmpl).implementation()
        

    def DefaultTemplateRenderer(self, values, templatename = None):
        """
        Renders the default template configured in context.configuration.template 
        with the given dictionary `values`. Calls `CacheHeader` to set the default 
        cache headers.
        
        Adds the following values if not set ::
            
            {"context": self.context, "view": self} 
        
        Template lookup first searches the current view module and if not found
        the parent or extended view module. 
        """
        if not templatename:
            templatename = self.context.configuration.template
            if not templatename:
                templatename = self.context.configuration.id
        tmpl = self._LookupTemplate(templatename)
        if not tmpl:
            raise ConfigurationError("Template not found: %(name)s %(type)s." % {"name": templatename, "type": self.context.configuration.id})
        if not "context" in values: values["context"] = self.context
        if not "view" in values: values["view"] = self
        response = render_to_response(tmpl, values, request=self.request)
        self.CacheHeader(response, user=self.User())
        return response

    def _LookupTemplate(self, tmplfile):
        conf = self.configuration
        if not tmplfile:
            raise TypeError("'tmplfile' is None. Need a template name to lookup the path.")
        if not conf or ":" in tmplfile:
            return tmplfile
        path = conf.templates
        if not path.endswith((":","/")):
            path += "/"
        fn = path + tmplfile
        try:
            if get_renderer(fn):
                return fn
        except ValueError:
            pass
        if not conf.parent:
            return tmplfile
        path = conf.parent.templates
        if not path.endswith((":","/")):
            path += "/"
        fn = path + tmplfile
        return fn


    def RenderView(self, obj, name="", secure=True, raiseUnauthorized=False, codepage="utf-8"):
        """
        Render a view for the object.
        
        *name* is the name of the view to be looked up.
        
        If *secure* is true permissions of the current user are checked against the view. If the
        user lacks the necessary permissions and empty string is returned or if *raiseUnauthorized*
        is True HTTPForbidden is raised. 
        
        returns rendered result
        """
        # store original context to reset after calling render_view
        orgctx = self.request.context
        orgresp = self.request.response
        orgname = self.request.view_name

        self.request.context = obj
        self.request.response = Response()
        self.request.view_name = name
        try:
            value = render_view(obj, self.request, name, secure)
            if value is None:
                value = ""
        except HTTPNotFound:
            value = "Not found!"
        except HTTPForbidden:
            value = ""
        finally:
            self.request.context = orgctx
            self.request.response = orgresp
            self.request.view_name = orgname

        #return str(value, codepage) # todo [3] unicode
        return value

    
    def Assets(self, assets=None, ignore=None, viewModuleConfID=None, types=None):  
        """
        Renders a list of static ressources as html <script> and <link>.
        If assets is None the list of assets is looked up in the view module-configuration.
        Asset definitions use a identifier and an asset path like: ::
        
            assets = (
              ('jquery.js', 'nive_cms.cmsview:static/mods/jquery.min.js'),
              ('toolbox.css', 'nive_cms.cmsview:static/toolbox/toolbox.css'),
              ('jquery-ui.js', 'http://example.com/static/mods/jquery.min.js')
            )
                
        If for example jquery is already included in the main page Assets() can be told to ignore certain
        entries: ::
        
            cmsview.Assets(ignore=["jquery.js"])  
            
        The list of assets is configured as part of the view module settings. If the view module 
        configuration is not stored as class attribute (`self.viewModule`) the configuration can be accessed 
        by id in the applications registry. In this case the id can be passed in like 
        `Assets(viewModuleConfID='editor')` to lookup the the configuration in the registry. 
        
        By default `Assets` renders both css and js links. You can use the parameter types e.g *types="js"*
        to get js file links only. Set *types="css"* to get all css links.
        """
        if viewModuleConfID:
            app = self.context.app
            conf = app.QueryConfByName(IViewModuleConf, viewModuleConfID)
            if not conf:
                return ""
            assets = conf.assets
        if assets is None and self.configuration:
            assets = self.configuration.assets
        if not assets:
            return ""
        
        if ignore is None:
            ignore = []
        
        js_tags = css_tags = []
        if types in (None, "js"):
            js_links = [self.StaticUrl(r[1]) for r in [v for v in assets if v[0] not in ignore and v[1].endswith(".js")]]
            js_tags = ['<script src="%s" type="text/javascript"></script>' % link for link in js_links]
        if types in (None, "css"):
            css_links = [self.StaticUrl(r[1]) for r in [v for v in assets if v[0] not in ignore and v[1].endswith(".css")]]
            css_tags = ['<link href="%s" rel="stylesheet" type="text/css" media="all">' % link for link in css_links]
        return ("\r\n").join(js_tags + css_tags)
        

    def tmpl(self):
        """
        Default function for template rendering.
        """
        return {}
    
    
    # file handling ------------------------------------------------------------------
    
    def file(self):
        return self.File()

    def File(self):
        """
        Used by "file" view for the current context. 
        
        Calls SendFile() for the file matching the filename of the current url.
        """
        if not len(self.request.subpath):
            raise NotFound
        path = self.request.subpath[0]
        if path in self.context.files:
            file = self.context.GetFile(path)
        else:
            file = self.context.GetFileByName(path)
        if not file:
            raise NotFound
        try:
            return self.SendFile(file)
        except FileNotFound:
            raise NotFound

    def SendFile(self, file, filename=None, headers=None):
        """
        Creates the response and sends the file back. Uses the FileIterator.
        
        #!date format
        """
        if not file:
            return HTTPNotFound()
        last_mod = file.mtime
        if not last_mod:
            last_mod = self.context.meta.pool_change
        r = Response(content_type=str(GetMimeTypeExtension(file.extension)), conditional_response=True, headerlist=headers)
        iterator = file.iterator()
        if iterator:
            r.app_iter = iterator
        else:
            try:
                r.body = file.read()
            except FileNotFound:
                raise NotFound
        r.content_length = file.size
        r.last_modified = last_mod
        r.etag = '%s-%s' % (last_mod, hash(file.path))
        r.cache_expires(self.fileExpires)
        if filename:
            r.content_disposition = 'attachment; filename=%s' % (filename)
        return r


    # http caching ----------------------------------------------------------------

    def CacheHeader(self, response, user=None):
        """
        Adds a http cache header to the response. If user is not None *NoCache()* is called, otherwise
        *Modified()*.
        
        returns response
        """
        if user is not None:
            return self.NoCache(response)
        return self.Modified(response, user)

    def NoCache(self, response, user=None):
        """
        Adds a no-cache header to the response.
        
        returns response
        """
        response.cache_control = "no-cache"
        response.etag = '%s-%s' % (response.content_length, str(self.context.id))
        return response

    def Modified(self, response, user=None):
        """
        Adds a last modified header to the response. meta.pool_change is used as date.

        returns response
        """
        if user:
            response.last_modified = formatdate(timeval=None, localtime=True, usegmt=True)
        else:
            t = self.context.meta.get("pool_change")
            if t and isinstance(t, str):
                t = ConvertToDateTime(t).timetuple()
                t = time.mktime(t)
            elif t:
                t = time.mktime(t.timetuple())
            else:
                t = None
            response.last_modified = formatdate(timeval=t, localtime=True, usegmt=True)
        response.etag = '%s-%s-%s' % (response.last_modified, response.content_length, str(self.context.id))
        return response


    # configuration -------------------------------------------------

    def GetViewConf(self, view_name=None):
        """
        Looks up the current view configuration `nive.definitions.ViewConf`.
        Please note: Unique view names on view module level are required for this
        function!

        view_name: If None the current view_name is used.
        returns `nive.definitions.ViewConf` or None
        """
        c = self.configuration
        if c is None:
            return None
        name = view_name or self.request.view_name
        found = []
        for v in c.views:
            if v.name==name:
                found.append(v)
        if len(found)==1:
            return found[0]
        elif len(found)==0:
            return None
        # try to find the right one if multiple names match
        #for v in found:
        #    if v.custom_predicates:
        #        for c in v.custom_predicates:
        #            # match the first one
        #            if c(*(self.context, self.request)):
        #                return v
        return None


    # user and security -------------------------------------------------
    
    @property
    def user(self):
        return self.User()
    
    def User(self, sessionuser=True):
        """
        Get the currently signed in user. If sessionuser=False the function will return
        the uncached write enabled user from database.
        
        returns the *Authenticated User Object* or None
        """
        return User(self.context, self.request, sessionuser)

    def UserName(self):
        """
        returns the *Authenticated User Name* or None
        """
        return authenticated_userid(self.request)    

    def Allowed(self, permission, context=None):
        """
        Check if the current user has the *permission* for the *context*.
        If context is none the current context is used
        
        returns True / False
        """
        # fix missing callback principals() context param
        context = context or self.context
        if context != self.request.context:
            try:
                originalContext = self.request.context
                self.request.context = context
                return self.request.has_permission(permission, context or self.context)
            finally:
                self.request.context = originalContext

        return self.request.has_permission(permission, context or self.context)

    def InGroups(self, groups):
        """
        Check if current user is in one of the groups.
        """
        user = self.User()
        if not user:
            return False
        return user.InGroups(groups)
    

    # render helper ---------------------------------------------------------
    
    def GetLocale(self):
        if hasattr(self, "_c_locale"):
            return self._c_locale
        l = "en"
        self._c_locale = l
        return l
    
    def GetTimezone(self):
        if hasattr(self, "_c_tz"):
            return self._c_tz
        l = "GMT"
        self._c_tz = l
        return l
    
    def RenderField(self, fld, data=None, context=None, user=None, htmlcode=True):
        """
        Render the data field for html display. Rendering depends on the datatype defined
        in the field configuration.
        
        If data is None the current context is used.
        
        returns string
        """
        if context is None:
            context = self.context
        if isinstance(fld, str):
            fld = context.GetFieldConf(fld)
        if not fld:
            return _("<em>Unknown field</em>")
        if data is None:
            data = context.data.get(fld['id'], context.meta.get(fld['id']))
        if fld['datatype']=='file':
            url = self.FileUrl(fld['id'])
            if not htmlcode:
                return url
            if not url:
                return ""
            url2 = url.lower()
            if url2.find(".jpg")!=-1 or url2.find(".jpeg")!=-1 or url2.find(".png")!=-1 or url2.find(".gif")!=-1:
                return """<img src="%s">""" % (url)
            return """<a href="%s">download</a>""" % (url)
        return FieldRenderer(context).Render(fld, data, context=context, user=user or self.user)
    

    def Translate(self, text):
        """
        Tranlate a translation string. Extracts language from request.
        """
        localizer = get_localizer(self.request)
        return localizer.translate(text)
        

    @property
    def utilities(self):
        """
        You can easily make utility functions like renderers and format functions
        accessible in templates by adding the functions to your view module
        configuration. e.g. ::

            def renderTime(value):
                return value.strftime("%H:%M")

            view_configuration = ViewModuleConf(
                id = "views",
                name = "Views",
                # utilities
                utilities = Conf(renderTime=renderTime)
                ...
            )

        You can call `renderTime` in your template by accessing the view class instance ::

            view.utilities.renderTime(value)


        :return: utilities mapping object
        """
        return self.configuration.get("utilities")

    # parameter handling ------------------------------------------------------------

    def GetFormValue(self, key, default=None, request=None, method=None):
        """
        Extract a form field from request. 
        Works with webob requests and simple dictionaries.
        
        returns the value or *default*
        """
        if not request:
            request = self.request
        if method is None and hasattr(request, "method"):
            method = request.method
        try:
            if method == "POST":
                if request.content_type=="application/json":
                    value = request.json_body.get(key)
                else:
                    value = request.POST.getall(key)
            elif method == "GET":
                value = request.GET.getall(key)
            else:
                if request.content_type=="application/json":
                    value = request.json_body.get(key)
                else:
                    value = request.POST.getall(key)
                if not value:
                    value = request.GET.getall(key)
            if isinstance(value, bytes):
                value = str(value, self.context.app.configuration.frontendCodepage)
        except (AttributeError,KeyError):
            if method == "POST":
                if request.content_type=="application/json":
                    try:
                        value = request.json_body.get(key)
                    except AttributeError:
                        value = None
                else:
                    value = request.POST.get(key)
            elif method == "GET":
                value = request.GET.get(key)
            else:
                if request.content_type=="application/json":
                    value = request.json_body.get(key)
                else:
                    value = request.POST.get(key)
                if not value:
                    value = request.GET.get(key)
            if isinstance(value, bytes):
                value = str(value, self.context.app.configuration.frontendCodepage)
            if value is None:
                return default
            return value
        if not len(value):
            return default
        elif len(value)==1:
            if value[0] is None:
                return default
            return value[0]
        if value is None:
            return default
        return value            


    def GetFormValues(self, request=None, method=None):
        """
        Extract all form fields from request. 
        Works with webob requests and simple dictionaries.
        
        returns dictionary
        """
        if not request:
            request = self.request
        if method is None and hasattr(request, "method"):
            method = request.method
        try:
            if method == "POST":
                if request.content_type=="application/json":
                    values = request.json_body
                else:
                    values = request.POST.mixed()
            elif method == "GET":
                values = request.GET.mixed()
            else:
                values = request.GET.mixed()
                if request.content_type=="application/json":
                    values.update(request.json_body)
                else:
                    values.update(request.POST.mixed())
        except AttributeError:
            try:
                if method == "POST":
                    if request.content_type=="application/json":
                        values = request.json_body
                    else:
                        values = request.POST
                elif method == "GET":
                    values = request.GET
                else:
                    values = {}
                    values.update(request.GET)
                    if request.content_type=="application/json":
                        values.update(request.json_body)
                    else:
                        values.update(request.POST)
            except AttributeError:
                values = request
        if not values:
            return {}
        return values
    

    def FmtURLParam(self, **kw):
        """
        Format all kw items as url parameter. 
        
        returns string
        """
        url = []
        params = kw
        #opt
        for p in list(params.keys()):
            if len(url):
                url.append("&")
            url.append("%s=%s" % (p, params[p]))
        return "".join(url)    #urllib.quote_plus()
    

    def FmtFormParam(self, **kw):
        """    
        Format all kw items as hidden input fields for forms.
        
        returns string 
        """
        form = []
        params = kw
        for p in list(params.keys()):
            value = ConvertToStr(params[p])
            form.append("<input type='hidden' name='%s' value='%s'>" % (p, value))
        return "".join(form)


    def FmtTextAsHTML(self, text):
        """
        Converts newlines to <br>.

        returns string
        """
        if not text:
            return ""
        if text.find("\r\n")!=-1:
            text = text.replace("\r\n","<br>\r\n")
        else:
            text = text.replace("\n","<br>\n")
        return text

    def FmtDateText(self, date, language=None):
        """
        Formats dates as readable text in conventional format.

        returns string
        """
        if not date:
            return ""
        if not isinstance(date, datetime):
            date = ConvertToDateTime(date)
        return date.strftime("%c")

    def FmtDateNumbers(self, date, language=None):
        """
        Formats dates as numbers e.g 13.12.2011.

        returns string
        """
        if not date:
            return ""
        if not isinstance(date, datetime):
            date = ConvertToDateTime(date)
        return date.strftime("%x")

    def FmtSeconds(self, secs):
        """ seconds to readable text """
        return FmtSeconds(secs)

    def FmtBytes(self, size):
        """ bytes to readable text """
        return FormatBytesForDisplay(size)

    def FmtTag(self, tag, closeTag="tag", **attributes):
        """
        Write a html tag with attributes.
        closeTag: 'tag', 'inline', 'no', 'only'
        """
        if closeTag=="only":
            return "</%s>" % (tag)
        attrs = ""
        if attributes:
            attrs = ['%s="%s"'%(a[0],str(a[1])) for a in list(attributes.items())]
            attrs = " " + " ".join(attrs)
        if closeTag=="inline":
            return "<%s%s/>" % (tag, attrs)
        elif closeTag in (None,'no'):
            return "<%s%s>" % (tag, attrs)
        return "<%s%s></%s>" % (tag, attrs, tag)

    def CutText(self, text, length):
        """ bytes to readable text """
        return CutText(text, length)


    def mark(self):
        return time.time() - self._t

    def mark2(self):
        try:
            return """<div class="mark">%.04f</div>""" % (time.time() - self.request.environ.get("START_TIME", self._t))
        except:
            return ""


    def HtmlTitle(self):
        t = self.request.environ.get("htmltitle")
        if t:
            return t
        return self.context.GetTitle()



# Configuration helper -------------------------------------------------
def UpdateViewConf(viewModule, viewname, **viewconfig):
    """
    Update and replace the view configuration for the view `viewname` in the given view module.
    You can pass a single or multiple values as `viewconfig` to overwrite the original views
    configuration.

    :param viewModule:
    :param viewname:
    :param **viewconfig:
    :return: None
    """
    p = 0
    for v in viewModule.views:
        if v.name==viewname:
            # create a copy fo the view and replace the original
            vcopy = ViewConf(copyFrom=v, **viewconfig)
            viewModule.views = list(viewModule.views)
            viewModule.views[p] = vcopy
            viewModule.views = tuple(viewModule.views)
            return
        p += 1



# Response utilities -------------------------------------------------
from zope.interface import implementer
from pyramid.interfaces import IExceptionResponse

@implementer(IExceptionResponse)
class ExceptionalResponse(HTTPOk):
    def __init__(self, headers=None, **kw):
        # HTTPException: map status int to self.code
        if "status" in kw:
            status = kw["status"]
            if status is not None:
                self.code = int(kw["status"].split()[0])
            del kw["status"]
        super(HTTPOk,self).__init__(headers=headers, detail=kw.get("detail","HTTP OK Response"), **kw)

    def __str__(self):
        return self.detail

    def __call__(self, environ, start_response):
        return Response.__call__(self, environ, start_response)


def SendResponse(data, mime="text/html", filename=None, raiseException=True, status=None, headers=None):
    """
    See views.BaseView class function for docs
    """
    cd = None
    if filename:
        cd = 'attachment; filename=%s'%(filename)
    if raiseException:
        raise ExceptionalResponse(content_type=mime, body=data, content_disposition=cd, status=status, headers=headers, charset="utf-8") # todo [3] unicode
    return Response(content_type=mime, body=data, content_disposition=cd, status=status, headerlist=headers, charset="utf-8") # todo [3] unicode


def Redirect(url, request, messages=None, slot="", raiseException=True, refresh=True, force=False):
    """
    See views.BaseView class function for docs
    """
    if not force and request.environ.get("HTTP_X_REQUESTED_WITH")=="XMLHttpRequest":
        # ajax call -> use relocate to handle this case
        return Relocate(url, request, messages, slot, raiseException, refresh)
    if messages:
        if isinstance(messages, str):
            request.session.flash(messages, slot)
        else:
            for m in messages:
                request.session.flash(m, slot)
    headers = None
    if hasattr(request.response, "headerlist"):
        headers = request.response.headerlist
    if raiseException:
        raise HTTPFound(location=url, headers=headers)
    return HTTPFound(location=url, headers=headers)
    

def Relocate(url, request, messages=None, slot="", raiseException=True, refresh=True):
    """
    See views.BaseView class function for docs
    """
    if messages:
        if isinstance(messages, str):
            request.session.flash(messages, slot)
        else:
            for m in messages:
                request.session.flash(m, slot)
    #if isinstance(url, str): # todo [3] unicode ?
    #    url = url.encode("utf-8")
    body = json.dumps({"location": url, "messages": messages, "refresh": refresh})
    headers = [('X-Relocate', url)]
    if hasattr(request.response, "headerlist"):
        for h in list(request.response.headerlist):
            # remove content type and content length
            if h[0].lower() in ("content-type", "content-length"):
                continue
            headers.append(h)
    if raiseException:
        raise ExceptionalResponse(headers=headers, body=body, content_type="application/json; charset=utf-8")
    return body


def ResetFlashMessages(request, slot=""):
    """
    See views.BaseView class function for docs
    """
    while request.session.pop_flash(slot):
        continue
        
    
def AddHeader(request, name, value, append=False):
    """
    See views.BaseView class function for docs
    """
    if append:
        headers = [(name, value)]
        if hasattr(request.response, "headerlist"):
            headers += list(request.response.headerlist)
        request.response.headerlist = headers
    else:
        request.response.headers[name] = value


def PreflightRequest(request,
                     allowOrigins="*",
                     allowMethods="POST, GET, DELETE, OPTIONS",
                     allowHeaders="",
                     allowCredentials=True,
                     maxAge=3600,
                     response=None):
    """
    Handles Access-Control preflight requests.

    :param request: The current request.
    :param allowOrigins: Either '*' (all allowed) or a list of hosts e.g ('http://mydomain.com')
    :param allowMethods: 'POST, GET, DELETE, OPTIONS'. Allowed http methods as string.
    :param allowHeaders: Custom headers passed in a request as string.
    :param allowCredentials: True/False. Credentials are made accessible in the client application.
    :param response: The response to be returned. if None `request.response` is used.
    :return: response
    """
    response = response or request.response
    origin = request.headers.get("Origin")
    if origin is None:
        # not a access-control preflight?
        return response
    if allowOrigins!="*":
        # check request origin, might also be checked against host
        if not origin in allowOrigins and not "*" in allowOrigins:
            # not allowed
            # set status to 405
            response.status="405 Not Allowed"
            return response
    headers = [
        ("Access-Control-Allow-Origin", origin),
        ("Access-Control-Allow-Methods", allowMethods),
        ("Access-Control-Allow-Credentials", str(allowCredentials).lower()),
        ("Access-Control-Max-Age", maxAge),
        ("Access-Control-Allow-Headers", allowHeaders),
        ("Vary", "Accept-Encoding, Origin")
    ]
    if hasattr(response, "headerlist"):
        headers += list(response.headerlist)
    response.headerlist = headers
    return response


def OriginResponse(request,
                   allowOrigins="*",
                   allowCredentials=True,
                   exposeHeaders="",
                   response=None):
    """
    Handles Access-Control responses to Origin header containing requests. This function
    adds required headers only and does not change the reponse.

    Returns None if the Origin does not match or the header is not present in the request.

    :param request: The current request.
    :param exposeHeaders: A list of custom headers to be made accessible in the client application as string.
    :param allowCredentials: True/False. Credentials are made accessible in the client application.
    :param response: The response to be returned. if None `request.response` is used.
    :return: response or None
    """
    response = response or request.response
    origin = request.headers.get("Origin")
    if origin is None or allowOrigins is None:
        # not a access-control aware request?
        return response
    if allowOrigins!="*":
        # check request origin, might also be checked against host
        if not origin in allowOrigins and not "*" in allowOrigins:
            # not allowed
            return None
    headers = [
        ("Access-Control-Allow-Origin", origin),
        ("Access-Control-Allow-Credentials", str(allowCredentials).lower()),
        ("Access-Control-Expose-Headers", exposeHeaders),
        ("Vary", "Accept-Encoding, Origin")
    ]
    if hasattr(response, "headerlist"):
        headers += list(response.headerlist)
    response.headerlist = headers
    return response


def User(context, request, sessionuser=True):
    """
    Get the currently signed in user. If sessionuser=False the function will return
    the uncached write enabled user from database.

    returns the *Authenticated User Object* or None
    """
    # cached session user object
    if not sessionuser:
        ident = authenticated_userid(request)
        if not ident:
            return None
        return context.app.portal.userdb.root.LookupUser(ident=ident)
    try:
        user = request.environ["authenticated_user"]
        if user:
            return user
    except (AttributeError, KeyError):
        pass
    ident = authenticated_userid(request)
    if not ident:
        return None
    return context.app.portal.userdb.root.GetUser(ident)


class FieldRenderer(object):
    
    def __init__(self, context, skip=()):
        self.context = context
        self.skipRender = skip
        
    def Render(self, fieldConf, value, useDefault=False, listItems=None, context=None, **kw):
        """
        fieldConf = FieldConf of field to be rendered
        value = the value to be rendered
        useDefault = use default values from fieldConf if not found in value
        listItems = used for list lookup key -> name
        context = context object the field is rendered for. Required if listItems=None.
        
        **kw:
        static = static root path for images
        """
        data = ""
        if useDefault:
            data = fieldConf["default"]
        if value is not None:
            data = value
        if fieldConf["id"] in self.skipRender:
            return data

        def loadListItems(fld, context, user=None):
            if not context:
                return []
            pool_type = context.GetTypeID() 
            return helper.LoadListItems(fld, app=context.app, obj=context, pool_type=pool_type, user=user)
        
        # format for output
        fType = fieldConf["datatype"]

        # fomat settings
        settings = fieldConf.get("settings",{})
        if settings:
            fmt = settings.get("format")
            if fmt=="bytesize":
                data = FormatBytesForDisplay(data)
                return data
            elif fmt=="image":
                tmpl = settings.get("path", "")
                path = tmpl % {"data":data, "static": kw.get("static","")}
                data = """<img src="%(path)s" class="img-fluid" title="%(name)s">""" % {"path":path, "name": fieldConf.name}
                return data
        
        if fType == "bool":
            if data:
                data = _("Yes")
            else:
                data = _("No")

        elif fType == "string":
            if settings.get("relation") == "userid":
                # load user name from database
                try:
                    udb = context.app.portal.userdb.root
                    user = udb.GetUser(data, activeOnly=0)
                    if user:
                        data = user.GetTitle()
                except AttributeError:
                    pass

        elif fType == "text":
            fmt = settings.get("format")
            if fmt=="newlineToBr":
                data = data.replace("\r\n", "\r\n<br>")
            elif fmt=="markdown":
                import markdown2
                data = markdown2.markdown(data)

        elif fType == "code":
            import html
            data = html.escape( data, False )

        elif fType == "date":
            if not isinstance(data, datetime):
                if not data:
                    return ""
                data = ConvertToDateTime(data)
            fmt = settings.get("strftime", "%x")
            return data.strftime(fmt)

        elif fType in ("datetime", "timestamp"):
            fmt = settings.get("format")
            if not isinstance(data, datetime):
                if not data:
                    return ""
                data = ConvertToDateTime(data)
            # defaults
            fmt = settings.get("strftime")
            if not fmt:
                fmt = "%x %H:%M"
                # hide hour and minutes if zero
                if data.hour==0 and data.minute==0 and data.second==0:
                    fmt = "%x"
                elif settings.get("seconds"):
                    fmt = "%x %X"
            return data.strftime(fmt)

        elif fType == "unit":
            if data == 0 or data == "0":
                return ""
            return data

        elif fType == "unitlist":
            if isinstance(data, (list, tuple)):
                return ConvertListToStr(data)
            return str(data)

        elif fType == "list" or fType == "radio":
            options = listItems or loadListItems(fieldConf, context, user=kw.get("user"))
            if not options:
                options = fieldConf.get("listItems")
                if hasattr(options, "__call__"):
                    options = options(fieldConf, self.context, user=kw.get("user"))

            if options:
                for item in options:
                    if item["id"] == data:
                        data = item["name"]
                        break
                
        elif fType in("multilist", "checkbox", "mselection", "mcheckboxes"):
            values = []
            options = listItems or loadListItems(fieldConf, context, user=kw.get("user"))
            if not options:
                options = fieldConf.get("listItems")
                if hasattr(options, "__call__"):
                    options = options(fieldConf, self.context, user=kw.get("user"))

            if isinstance(data, str):
                data = tuple(data.split("\n"))
            if isinstance(data, (float,int)):
                data = str(data)
            for ref in data:
                if options:
                    for item in options:
                        if item["id"] == ref:
                            values.append(item["name"])
                else:
                    values.append(ref)
            delimiter = ", "
            if settings and "delimiter" in settings:
                delimiter = settings["delimiter"]
            data = delimiter.join(values)

        elif fType == "url":
            if render:
                if data != "" and data.find("http://") == -1:
                    data = "http://" + data
                l = data[7:50]
                if len(data) > 50:
                    l += "..."
                data = "<a alt='%s' href='%s' target='_blank'>%s</a>" % (data, data, l)

        elif fType == "urllist":
            urllist = data
            data = []
            if len(urllist) and urllist[0] == "[":
                urllist = urllist[1:-1]
                ul = urllist.split("', '")
                ul[0] = ul[0][1:]
                ul[len(ul)-1] = ul[len(ul)-1][:-1]
            else:
                ul = urllist.replace("\r","")
                ul = ul.split("\n")
            links = []
            for l in ul:
                t = l.split("|||")
                title = ""
                url = t[0]
                if len(t) > 1:
                    title = t[1]
                if title != "":
                    title = '%s (%s)' % (title, url)
                else:
                    title = url
                links.append({"id": l, "name": title})
                data.append("<a alt='%s' title='%s' href='%s' target='_blank'>%s</a><br>" % (title, title, url, title))
            data = "".join(data)

        elif fType == "password":
            data = "*****"

        elif fType == "lines":
            if data:
                tmpl = "%s<br>"
                if settings and "tmpl" in settings:
                    tmpl = settings["tmpl"]
                if isinstance(data, str):
                    data = data.split("\n")
                data = "".join([tmpl%ll for ll in data])

        elif fType == "json":
            if data:
                data = json.dumps(data)

        return data
    

# Mail template mapper -------------------------------------------------

class Mail(object):
    """
    Mail template object with template link and title. ::
    
        mail = Mail(title="New mail", tmpl="mypackage:templates/foo.pt")
        body = mail(value1=1, value2=2)
        
    the body of the mail is generated by calling the mail object. *kw* parameters
    are passed on in template.render()
    """
    title = ""
    tmpl = None     # 'mypackage:templates/foo.pt'
    
    def __init__(self, title="", tmpl=None):
        self.title = title
        self.tmpl = tmpl
    
    def __call__(self, **kws):
        mail = render(self.tmpl, kws)
        return mail



class reify(object):

    """ Put the result of a method which uses this (non-data)
    descriptor decorator in the instance dict after the first call,
    effectively replacing the decorator with an instance variable."""

    def __init__(self, wrapped):
        self.wrapped = wrapped

    def __get__(self, inst, objtype=None):
        if inst is None:
            return self
        val = self.wrapped(inst)
        setattr(inst, self.wrapped.__name__, val)
        return val




# Copyright 2012, 2013 Arndt Droullier, Nive GmbH. All rights reserved.
# Released under GPL3. See license.txt
#

from pyramid.i18n import get_localizer
from pyramid.threadlocal import get_current_request

from nive.tool import Tool, ToolView
from nive.helper import FakeLocalizer
from nive.definitions import ToolConf, FieldConf, IApplication
from nive.definitions import ViewConf
from nive.i18n import _

from nive.utils.utils import FormatBytesForDisplay

configuration = ToolConf(
    id = "cmsstatistics",
    context = "nive.tools.cmsstatistics.cmsstatistics",
    name = _("CMS Statistics"),
    description = _("This function provides a short summary of elements and data contained in the website."),
    apply = (IApplication,),
    mimetype = "text/html",
    data = [
        FieldConf(id="tag", datatype="string", default="cmsstatistics", hidden=1)
    ],
    views = [
        ViewConf(name="", view=ToolView, attr="form", permission="system", context="nive.tools.cmsstatistics.cmsstatistics")
    ]
)


class cmsstatistics(Tool):
    """
    """

    def _Run(self, **values):

        try:
            localizer = get_localizer(get_current_request())
        except:
            localizer = FakeLocalizer()

        self.InitStream()
        app = self.app
        datapool = app.db
        conn = datapool.connection
        c = conn.cursor()

        self.stream.write("<table class='table table-bordered'>\n")

        sql = "select count(*) from pool_meta"
        c.execute(sql)
        rec = c.fetchall()
        self.stream.write(localizer.translate(_("<tr><th>Elements in total</th><td>${value}</td></tr>\n", mapping={"value": rec[0][0]})))

        sql = "select count(*) from pool_files"
        c.execute(sql)
        rec = c.fetchall()
        self.stream.write(localizer.translate(_("<tr><th>Physical files</th><td>${value}</td></tr>\n", mapping={"value": rec[0][0]})))

        sql = "select sum(size) from pool_files"
        c.execute(sql)
        rec = c.fetchall()
        self.stream.write(localizer.translate(_("<tr><th>Physical files size</th><td>${value}</td></tr>\n", mapping={"value": FormatBytesForDisplay(rec[0][0])})))

        for t in app.configurationQuery.GetAllObjectConfs():
            sql = "select count(*) from pool_meta where pool_type='%s'" % t.id
            c.execute(sql)
            rec = c.fetchall()
            self.stream.write(localizer.translate(_("<tr><th>${name}</th><td>${value}</td></tr>\n", mapping={"name": t.name, "value": rec[0][0]})))
        
        self.stream.write("</table>\n")

        c.close()
        return self.stream, 1






import copy

from nive.utils.dataPool2.mysql.tests import test_MySql
try:
    from nive.utils.dataPool2.mysql.mySqlPool import *
except:
    pass

from . import test_db
from nive.utils.dataPool2.sqlite.sqlite3Pool import *


def create_mysql(n):
    pool = MySql(test_MySql.conf)
    pool.SetStdMeta(copy.copy(test_MySql.stdMeta))
    pool.GetPoolStructureObj().SetStructure(test_MySql.struct)
    pool.CreateConnection(test_MySql.conn)

    print("Create", n, "entries (data+meta+file): ",)
    t = time.time()
    for i in range(0,n):
        e=pool.CreateEntry("data1")
        if i==0:  id = e.GetID()
        e.data.update(test_MySql.data1_1)
        e.meta.update(test_MySql.meta1)
        e.CommitFile("file1", {"file":test_db.file1_1, "filename":"file1.txt"})
        #e.Commit()
    t2 = time.time()

    pool.Close()
    print(n, " entries in ", t2-t, "secs. ", (t2-t)/n, " per entry")
    print()
    return id


def create_sqlite3(n):
    pool = Sqlite3(test_db.conf)
    pool.SetStdMeta(copy.copy(test_db.stdMeta))
    pool.GetPoolStructureObj().SetStructure(test_db.struct)
    pool.CreateConnection(test_db.conn)

    print("Create", n, "entries (data+meta+file): ",)
    t = time.time()
    for i in range(0,n):
        e=pool.CreateEntry("data1")
        if i==0:  id = e.GetID()
        e.data.update(test_db.data1_1)
        e.meta.update(test_db.meta1)
        e.CommitFile("file1", {"file":test_db.file1_1, "filename":"file1.txt"})
        e.Commit()
    t2 = time.time()

    pool.Close()
    print(n, " entries in ", t2-t, "secs. ", (t2-t)/n, " per entry")
    print()
    return id



# ----------------------------------------------------------------------------------------------

def run():
    n = 1 * 500
    #id = create_mysql(n)
    id = create_sqlite3(n)


if __name__ == '__main__':
    run()




import sys
import traceback

from pyramid.exceptions import Forbidden
from pyramid.httpexceptions import HTTPError
from pyramid.renderers import render_to_response

from nive import portal


class IfacePortal(portal.Portal):

    # TODO remove default views
    def SetupPortalViews(self, config):
        # redirects
        config.add_view(servererror, context=Exception)
        config.add_view(httperror, context=HTTPError)
        config.add_view(portal.forbidden_view, context=Forbidden)
        config.add_view(portal.portal_view, name="", context="nive.portal.Portal")
        config.add_view(portal.logout_view, name="logout", context="nive.portal.Portal")
        config.add_view(portal.login_view,  name="login", context="nive.portal.Portal")
        config.add_view(portal.account_view,name="account", context="nive.portal.Portal")
        config.add_view(portal.robots_view, name="robots.txt", context="nive.portal.Portal")
        config.add_view(portal.sitemap_view, name="sitemap.xml", context="nive.portal.Portal")
        #config.add_view(favicon_view, name="favicon.txt", context="nive.portal.Portal", view=PortalViews)
    
        config.commit()


# http errors and redirects ------------------------------------------------------------------

def servererror(errorResponse, request):
    """
    Called in case of server side exceptions resulting in 500 status codes
    """
    if not isinstance(errorResponse, str):
        errorResponse = str(errorResponse)
    resp = render_to_response("nive.components.iface:templates/error.pt",
                              {"error":"500", "message":"The error was: %s" % errorResponse},
                              request=request)
    resp.status = "500 Server error"
    traceback.print_exc(file=sys.stdout)
    return resp

def httperror(errorResponse, request):
    """
    Queries the context object to find out whether to redirect or return the error status only.
    
    1) request.context.ErrorType(request.view_name)==1 -> Status response
    2) request.environ["HTTP_X_REQUESTED_WITH"]==XMLHttpRequest -> Status response
    3) all other cases -> default error page
    
    Extracts the context and the current url from the workspace, looks up the workspace runtime 
    configuration and redirects to the configured forbidden url.
    """
    response = errorResponse
    def status():
        headers = [("X-Result", "false")]
        if hasattr(response, "headerlist"):
            headers += list(response.headerlist)
        response.headerlist = headers    
        return response
    # 1)
    try:
        if request.context.HTTPErrorType(request.view_name)==1:
            return status()
    except:
        pass
    # 2)
    if request.environ.get("HTTP_X_REQUESTED_WITH")=="XMLHttpRequest":
        return status()
    # 3)
    resp= render_to_response("nive.components.iface:templates/error.pt",
                              {"error":"404"},
                              request=request)
    resp.status = "404 Not found"
    return resp
 





from nive.definitions import Conf, IRoot, IContainer
from nive.i18n import translate, _


class Hint(object):
    """
    Hint class
    Use hints to point users to outstanding actions and explain the situation.
    
    Hints can linked be linked to hint-sections defined in the interface and be
    loaded on precondition callback functions. Callbacks take the current view
    context as parameter.
    
    The user can disable hints in his personal settings.  
    """
    
    def __init__(self, id, sections, condition, text):
        self.id = id
        self.sections = sections
        self.condition = condition
        self.text = text
        
    def Visible(self, section, view):
        if not section in self.sections:
            return False
        return self.condition(view)

    def Render(self, values=None):
        return self.text.render(values)



class Parts:
    """
    Renders interface parts
    """
    
    action = None
    static = ""
    activateUnitInfo = False
        
    # html parts -------------------------------------------------------

    def title(self):
        """
        html head title
        """
        t = self.context.app.configuration.title
        if t:
            return "%s - %s" % (self.ifaceConf.name, t)
        return self.ifaceConf.name

    @property
    def tmpl_macro(self):
        return self.ifaceConf.tmpl_macro

    def sections(self, count=100, context=None):
        """
        html body div#header div#headOptions div#mainNav
        
        main navigation list
        
        count: number of sections to be rendered. 
        """
        tmpl = """<li class="nav-item %(class)s"><a class="nav-link" href="%(url)s" title="%(description)s"> %(name)s </a></li>"""
        html = []
        try:
            n = self.request.currentSection
        except:
            try:
                n=self.ifaceConf.currentSection
            except:
                n = self.GetFormValue("sec")
        if self.ifaceConf.sections == "portal":
            self.ifaceConf.sections = self.context.app.portal.sections
        for s in self.ifaceConf.sections[:count]:
            permission = s.get("permission")
            if permission and not self.Allowed(permission, context or self.context):
                continue
            cls = "normal"
            if n == s["ref"]:
                cls = "active"
            url = self.ToUrl(s)
            html.append(tmpl % {"url": url, "class": cls, "name": translate(s["name"]), "description": translate(s.get("description",""))})
        return "".join(html)
    

    def backlink(self):
        """
        backlink added to head logo.
        """
        if not self.ifaceConf.backlink:
            return "/"
        return self.ToUrl(self.ifaceConf.backlink)


    def headlink(self):
        """
        single link in header right to headoptions drop down.
        """
        if not self.ifaceConf.headlink:
            return ""
        conf = self.ifaceConf.headlink
        if conf.get("icon"):
            return """
<a href="%(url)s" class="nav-link"><img src="%(static)s%(icon)s" title="%(name)s"> %(name)s</a>
        """ % {"url": self.ToUrl(conf), "name": translate(conf.name), "static": self.static, "icon": conf.icon}
        # no icon configured
        return """
<a href="%(url)s" class="nav-link">%(name)s</a>
        """ % {"url": self.ToUrl(conf), "name": translate(conf.name), "static": self.static}
        

    def headlinkAnonymous(self):
        """
        single link in header right to headoptions drop down.
        """
        if not self.ifaceConf.headlinkanonymous:
            return ""
        conf = self.ifaceConf.headlinkanonymous
        if conf.get("icon"):
            return """
<a href="%(url)s"><img src="%(static)s%(icon)s" title="%(name)s"> %(name)s</a>
        """ % {"url": self.ToUrl(conf), "name": translate(conf.name), "static": self.static, "icon": conf.icon}
        # no icon configured
        return """
<a href="%(url)s">%(name)s</a>
        """ % {"url": self.ToUrl(conf), "name": translate(conf.name), "static": self.static}


    def headoptions(self, context=None):
        """
        html body div#header div#userOptions
        user options in head: head_logout, head_search
        """
        html = []
        for u in self.ifaceConf.headoptions:
            if isinstance(u, str):
                if u=="divider":
                    html.append("""<div class="dropdown-divider"></div>""")
                else:
                    html.append(getattr(self, u)())
            else:
                permission = u.get("permission")
                if permission and not self.Allowed(permission, context or self.context):
                    continue
                html.append(self.head_link(u, cls="dropdown-item"))
        return "".join(html)
    

    def navigation(self):
        """
        left column navigation
        """
        tmpl = """
 <div onclick="%(url)s" class="nav-link disabled nav-header">%(name)s</div>
 %(content)s
        """ 
        html = ["""<nav class="nav flex-column">"""]
        for n in self.ifaceConf.navigation:
            html.append(tmpl % {"url": self.ToUrl(n), "content": getattr(self, n["widget"])(n), "name": translate(n["name"], self.request)})
        html.append("""</nav> """)
        return "".join(html)

        
    def path_render(self, items, css="path"):
        """
        Renders path items as html
        
        items: dict with url, title, class 
        """
        sep = """ <span class="divider">/</span>"""
        tmpl = """<li class="%(class)s"><a href="%(url)s">%(title)s</a></li>"""
        html=["""<ul class="%s">"""%css]
        html.append(sep.join([tmpl % i for i in items]))
        html.append("</ul>")
        return "".join(html)

    
    def path(self, object=None, css="path", addLink=None):
        """
        Parent links in content top
        addLink: custom last element in list as Conf(id='url',name='name') 
        
        use css=breadcrumb for bootstrap style
        """
        if not self.ifaceConf.path:
            return ""
        html=[]
        if addLink:
            html.append({"url":self.ToUrl(addLink.get('id')), "title":translate(addLink.get('name')), "class":""})
        if not object:
            o = self.context
        else:
            o = object
        objs = list(o.GetParents())
        objs.reverse()
        for p in objs:
            html.append({"url":self.FolderUrl(p), "title":p.GetTitle(), "class":""})
        html.append({"url":self.FolderUrl(o), "title":o.GetTitle(), "class":"active"})
        return self.path_render(html, css=css)
    
    
    def contentHeader(self, pageinfo=None):
        """
        renders the title for the content area
        """
        if pageinfo:
            return pageinfo.get("title", "")
        try:
            return self.context.meta.title
        except:
            return ""
    
    
    def tabs(self, active=None):
        """
        html body div.mainBoxContent div#unit_tabs
        tabs in content
        """
        tabs = self.GetTabs(self.context)
        if not tabs:
            return ""
        tmpl = """<li class="nav-item"><a class="nav-link %(cls)s" href="%(url)s">%(name)s</a></li>"""
        if active is None:
            try:
                active = self.request.currentTab
            except:
                pass
        html = ["""<ul class="nav nav-tabs">"""]
        for t in tabs:
            permission = t.get("permission")
            if permission and not self.Allowed(permission, self.context):
                continue
            cls = ""
            if active == t["id"]:
                cls = "active"
            html.append(tmpl % {"url": self.ToUrl(t), "name": translate(t["name"]), "cls": cls})
        html.append("</ul>")
        return "".join(html)


    def settingsLinks(self, base=None):
        """
        html body div.mainBoxContent div#unit_settings
        dropdown in content
        """
        li = self.settingsStripped(base=base)
        if not li:
            return ""
        return """
<div class="dropdown float-right unit-settings">
  <button class="btn btn-sm btn-light dropdown-toggle" data-toggle="dropdown" aria-haspopup="true" aria-expanded="false">
    <i class="icon-cog"></i> %(Settings)s
    <span class="caret"></span>
  </button>
  <div class="dropdown-menu" aria-labelledby="dropdownMenuButton">
    %(li)s
  </div>
</div>  """% dict(li=li, Settings=translate(_("Settings"), self.request))


    def settingsStripped(self, active=None, base=None, addSlots=False, cls="dropdown-item"):
        """
        settings as <li> list
        """
        settings = self.GetSettings(self.context)
        if addSlots:
            service = self.context.service
            if service is not None:
                for slot in service.slots:
                    settings.append(Conf(id="context.slot?sid="+slot, name=service.get(slot).name, icon=""))

        if not settings:
            return ""
        if active is None:
            try:
                active = self.request.currentTab
            except:
                pass

        app = self.context.app
        base0 = ""
        if base=="dev":
            from peak.config import ROUTING # TODO reconfigure ROUTING
            base = ROUTING.development
            base0 = self.Url(app.portal)
            if base.endswith("/") and not base0.endswith("/"):
                base = base[:-1]
            if not base.endswith("/") and base0.endswith("/"):
                base += "/"

        def rewriteBase(url):
            if not base0:
                return url
            return url.replace(base0, base)

        tmpl = """<a class="%(class)s" href="%(url)s">%(icon)s %(name)s</a>"""
        html = []
        for t in settings:
            permission = t.get("permission")
            if permission and not self.Allowed(permission, self.context):
                continue
            cc = cls
            if active == t["id"]:
                cc += " active"
            ic = t.get("icon","")
            icon = ("""<i class="%s"></i>"""%ic) if ic else ""
            html.append(tmpl % {"url": rewriteBase(self.ToUrl(t)), "name": translate(t["name"], self.request), "class": cc, "icon": icon})

        return "\r\n".join(html)


    def unitInfo(self, fields):
        """
        disabled!!!
        html body div.mainBoxContent div.unit_info
        block below tabs with object info
        """
        if not self.activateUnitInfo:
            return "" # TODO config unitinfo

        if self.context.IsRoot():
            return " "
        if fields == "edit":
            fields = (('pool_wfa', 'pool_create', 'pool_createdby'),('id', 'pool_change', 'pool_changedby'))
        elif type(fields) == type(""):
            fields = (('id', 'pool_state', 'pool_category'),)
        html = """
<div class="unit_info">
 <div class="blockOptions" style="float:right;margin:5px;">
  Details <a href="" id="unitInfo_b" onClick="javascript:swBl('unitInfo',this);return false;"><img src="%(static)simages/maximize.png" alt="Details ein-/ausblenden" style="vertical-align: bottom" /></a>
 </div>
 <table id="unitInfo" cellpadding="0" cellspacing="0">
  %(rows)s
 </table>
 <br style="clear:both"/>
 <!--script language="javascript">swBlLoad('unitInfo','unitInfo_b');</script-->
</div>
        """
        rows = []
        for row in fields:
            rows.append("<tr>")
            first = True
            for f in row:
                fld = self.context.GetFieldConf(f)
                cls = ""
                if first:
                    cls = " class='first'"
                    first = False
                rows.append("""<th%s>%s</th><td>%s</td>""" % (cls, fld["name"], self.RenderField(fld, self.context.meta.get(f, self.context.data.get(f)))))
            rows.append("</tr>")
        html = html % {"static": self.static, "rows": rows}
        return html

     
    def messages(self, msgs=None, error=0):
        """
        html body div.mainBoxContent ul
        displays messages if found in context or passed as msgs
        class="alert alert-success"
        """
        cls = "alert-success"
        if error:
            cls = "alert-error"
        if not msgs:
            msgs = self.GetFormValue("msgs")
        flashs = self.request.session.peek_flash()
        if not msgs and not flashs:
            return ""
        html = ["""<div class="alert %s"><ul>"""%cls]
        if msgs:
            for m in msgs:
                html.append("""<li>%s</li>""" % (m))
        if flashs:
            for m in self.request.session.pop_flash():
                html.append("""<li>%s</li>""" % (m))
        html.append("""</ul></div>""")
        return "".join(html)
    
    
    def hints(self, section):
        """
        Renders the given hint section for the current view context
        """
        if not self.ifaceConf.hints:
            return ""
        html = []
        for hint in self.ifaceConf.hints:
            if not hint.Visible(section, self):
                continue
            html.append("""<div class="alert hint">%s</div>"""%(str(hint.text)))
        return "".join(html)
     
     
    def shortcuts(self):
        """
        html body div.mainBoxContent div#unit_shortcuts
        shortcuts links / icons below unitInfo
        """
        sc = self.GetShortcuts(self.context)
        if not sc:
            return ""
        html = """<div id="unit_shortcuts" class="well alert alert-light"><div class="unit_options_block">Shortcuts</div> %s  <br class="clearBlock" /></div>"""
        l = []
        for s in sc:
            try:
                l.append(getattr(self, s["id"])())
            except Exception as e:
                l.append("Error %s (%s)" % (str(e), s["id"]))
                pass 
        return html % ("".join(l))
    
    
    # widgets ------------------------------------------------------
    """
    html widgets called in headoptions, navigation, shortcuts
    """
    
    def head_link(self, conf, cls=""):
        if conf.get("icon"):
            return """
<a href="%(url)s" class="%(cls)s"><img src="%(static)s%(icon)s" title="%(name)s"> %(name)s</a>
        """ % {"url": self.ToUrl(conf), "name": translate(conf.name), "static": self.static, "cls": cls, "icon": conf.icon}
        # no icon configured
        return """
<a href="%(url)s" class="%(cls)s">%(name)s</a>
        """ % {"url": self.ToUrl(conf), "name": translate(conf.name), "cls": cls, "static": self.static}

    def head_logout(self):
        """
        logout and help lilnk
        """
        return """
<li><a href="%(logout)s"><img src="%(static)simages/logout.png" title="Logout" align="top">%(username)s Logout</a></li>
        """ % {"logout": "/logout", "username": "", "static": self.static}


    def head_search(self):
        """
        simple searchform with textfield and button. calls search view.
        """
        return """
<div id="headSearch2">
 <span>Suche</span>
 <form action="%(url)ssearch'" method="get">
  <input type="text" name="searchPhrase" id="searchStr"/>
  <input type="submit" name="searchButton" id="searchButton" value="Start"/>
 </form>
</div>
        """ % {"url": self.FolderUrl(self.context.root)}
    
    
    def navigation_folder(self, conf):
        """
        javascript tree view for content structure.
        configuration: icon, sort, levelup, callback

        """
        sort = conf.get("sort", "title")
        container = self.context
        if not IContainer.providedBy(self.context):
            container = self.context.parent
        items = container.GetObjsList(parameter=dict(), sort=sort, containerOnly=True)
        if conf.get("icon"):
            tmpl = """<li class="nav-item"><a href="open?id=%(id)d" class="nav-link"><i class='""" + conf.get("icon") + """'></i> %(title)s</a></li>"""
        elif conf.get("state"):
            tmpl = """<li class="nav-item"><a href="open?id=%(id)d" class="nav-link">%(title)s [%(pool_state)s]</a></li>"""
        else:
            tmpl = """<li class="nav-item"><a href="open?id=%(id)d" class="nav-link">%(title)s</a></li>"""
        html = []
        if conf.get("levelup", True) and not IRoot.providedBy(self.context):
            html.append(tmpl % dict(id=container.parent.id, title="../ " + container.parent.meta.title, pool_state=""))
        cb = conf.get("callback")
        for i in items:
            if cb is not None:
                html.append(cb(i))
            else:
                html.append(tmpl % i)
        return "".join(html)


    def navigation_section(self, conf):
        """
        section with link list
        configuration: links 
        """
        tmpl = """<a href="%(url)s" class="nav-link">%(icon)s %(name)s</a>"""
        html = []
        for link in conf["links"]:
            icon = link.get("icon", "")
            if icon:
                icon = """<i class="%s"></i> """ % icon
            html.append(tmpl % {"url": self.ToUrl(link), "name": translate(link["name"], self.request), "icon": icon})
        return "".join(html)
    

    def navigation_path(self, conf):
        """
        renders the parent path as vertikal links
        """
        tmpl = """<li><a href="%(url)sview" class="level%(level)s">%(name)s</a></li>"""
        html = []
        object = self.context
        objs = list(object.GetParents())
        objs.reverse()
        objs += [self.context]
        l = 0
        for p in objs:
            html.append(tmpl % {"url":self.FolderUrl(p), "level": str(l), "name":p.GetTitle()})
            l+=1
        return "".join(html)
    

    def shortcut_add(self):
        """
        context sensitive list of add links
        """
        atypes = self.context.GetAllowedTypes(self.User())
        #atypes = self.context.app.GetAllObjectConfs(visibleOnly=True)
        if not atypes:
            return ""
        l = []
        for t in atypes:
            l.append("""<a href="add?pool_type=%(type)s"><img alt="%(name)s erstellen" src="%(static)simages/types/%(type)s_add.png" align="top" /> </a>""" % {"name": t["name"], "type": t["id"], "static": self.static})
        return """<div class="unit_options_block">%s</div>""" % ("".join(l))
        

    def shortcut_wfbuttons(self):
        """
        commit and reject icons if contained in context workflow transitions
        """
        if self.context.configuration["workflowDisabled"]:
            return ""
        wfinfo = self.context.wf.GetWfInfo(self.User())
        if not wfinfo or not wfinfo["wf"] or len(wfinfo["wf"]["transitions"])==0:
            return ""

        l = []
        redirect = self.GetFormValue('url')
        if not redirect:
            redirect = self.request.URL
        for t in wfinfo["transitions"]:
            a = ""
            if "commit" in t['applications']:
                a = "commit"
            if "reject" in t['applications']:
                a = "reject"
            if not a:
                continue
            p = self.FmtURLParam(action=a, trans=t['id'], url=redirect)
            l.append("""<a href="wf?%(param)s" class="tree"><img alt="%(name)s" src="%(static)simages/publish.png" align="top" /></a> """ % {"name": t["name"], "param": p, "static": self.static})
        return """<div class="unit_options_block">%s</div>""" % ("".join(l))
        

    def shortcut_wfselect(self):
        """
        drop down with all possible context workflow transitions
        """
        if self.context.configuration["workflowDisabled"]:
            return ""
        wfinfo = self.context.wf.GetWfInfo(self.User())
        if not wfinfo or not wfinfo["wf"] or len(wfinfo["wf"]["transitions"])==0:
            return ""

        l = []
        redirect = self.GetFormValue('url')
        if not redirect:
            redirect = self.request.URL
        for t in wfinfo["transitions"]:
            l.apend("""<option value="%s">%s</option>""" % (t['applications'][0], t["name"]))
        html = """
<div class="unit_options_block">
<form id="unit_options_transition" name="transition" action="wf" method="post">
 %(param)s
 <select name="action" onChange="submitdd(this)">
  <option>Status setzen...</option>
  %(options)s
 </select>
</form>
</div>
        """ % {"param": self.FmtFormParam(url=redirect), "options": "".join(l)}
        return html


    def shortcut_tools(self):
        """
        drop down with all possible context tools
        """
        tools = self.context.app.configurationQuery.GetAllToolConfs(self.context)
        if not tools:
            return ""
        redirect = self.GetFormValue("url")
        if not redirect:
            redirect = self.request.URL
        l = []
        for t in tools:
            l.append("""<option value="%s">%s</option>""" % (t["id"], t["name"]))
        html = """
<div class="unit_options_block">
<form id="unit_options_tool" action="tool" method="post">
 %(param)s
 <select name="tool" onChange="submitdd(this)">
  <option value="">Zusatzfunktionen...</option>
  %(tools)s
 </select>
</form>
</div>
        """ % {"param": self.FmtFormParam(redirect_url=redirect), "tools": "".join(l)}
        return html


    def shortcut_duplicate(self):
        """
        in place duplicate button
        """
        param = self.FmtURLParam(ids=self.context.id)
        l = """<a href="duplicate?%(param)s"><img src="%(static)simages/duplicate.png" alt="Duplizieren" align="top" /> </a>""" % {"param": param, "static": self.static}
        return """<div class="unit_options_block">%s</div>""" % (l)


    def shortcut_delete(self):
        """
        object delete button
        """
        if not self.Allowed("iface_delete", self.context):
            return ""
        p = self.FmtURLParam(ids=self.context.id)
        l = """<a href="../delete?%(param)s"><img src="%(static)simages/delete.png" alt="L&ouml;schen" align="top" /> </a>""" % {"param": p, "static": self.static}
        return """<div class="unit_options_block_del">%s</div>""" % (l)





# -*- coding: utf-8 -*-

from nive.components.adminview import view

from nive.tests.db_app import *
from nive.tests import __local

from pyramid import testing
from pyramid.renderers import render



class tViews(__local.DefaultTestCase):

    def setUp(self):
        request = testing.DummyRequest()
        request._LOCALE_ = "en"
        self.request = request
        self.request.content_type = ""
        self.config = testing.setUp(request=request)
        self.config.include('pyramid_chameleon')
        self._loadApp()
        self.portal = Portal()
        self.portal.Register(self.app, "nive")
        self.app.Register(view.configuration)
        self.app.Register(view.dbAdminConfiguration)
        self.app.Register("nive.components.reform")
        self.app.Startup(self.config)
        self.request.context = self.app
    

    def tearDown(self):
        self._closeApp()
        testing.tearDown()


    def test_basics(self):
        v = view.AdminBasics(context=self.app, request=self.request)
        self.assertTrue(v.GetAdminWidgets())
        self.assertTrue(v.RenderConf(view.configuration))
        self.assertTrue(v.RenderConf(view.dbAdminConfiguration))
        self.assertTrue(v.Format(view.configuration, view.configuration.id))
        self.assertTrue(v.Format(view.dbAdminConfiguration, view.configuration.id))
        self.assertTrue(v.AdministrationLinks(context=None))
        self.assertTrue(v.AdministrationLinks(context=self.app))


    def test_views(self):
        v = view.AdminView(context=self.app, request=self.request)
        v.__configuration__ = lambda: view.configuration
        v.view()
        self.assertTrue(v.index_tmpl())
        self.assertTrue(v.editbasics())
        self.assertTrue(v.editdatabase())
        self.assertTrue(v.editportal())
        self.assertTrue(v.tools())
        v.doc()


    def test_form(self):
        v = view.AdminBasics(context=self.app, request=self.request)
        form = view.ConfigurationForm(context=view.configuration, request=self.request, view=v, app=self.app)
        form.fields = (
            FieldConf(id="title",           datatype="string", size=255,  required=0, name="Application title"),
            FieldConf(id="description",     datatype="text",   size=5000, required=0, name="Application description"),
            FieldConf(id="workflowEnabled", datatype="bool",   size=2,    required=0, name="Enable workflow engine"),
            FieldConf(id="fulltextIndex",   datatype="bool",   size=2,    required=0, name="Enable fulltext index"),
            FieldConf(id="frontendCodepage",datatype="string", size=10,   required=1, name="Codepage used in html frontend"),
        )
        form.Setup()
        self.request.POST = {"name": "testuser", "email": "testuser@domain.net"}
        self.request.GET = {}

        self.assertTrue(form.Start(None))
        self.assertTrue(form.Update(None))
             


class tTemplates(__local.DefaultTestCase):

    def setUp(self):
        request = testing.DummyRequest()
        request._LOCALE_ = "en"
        self.request = request
        self.request.content_type = ""
        self.config = testing.setUp(request=request)
        self.config.include('pyramid_chameleon')
        self._loadApp()
        self.portal = Portal()
        self.portal.Register(self.app, "nive")
        self.app.Register(view.configuration)
        self.app.Register(view.dbAdminConfiguration)
        self.app.Register("nive.components.reform")
        self.app.Startup(self.config)
        self.request.context = self.app

    def tearDown(self):
        self.app.Close()
        testing.tearDown()

    def test_templates(self):
        v = view.AdminView(context=self.app, request=self.request)
        v.__configuration__ = lambda: view.configuration
        vrender = {"context":self.app, "view":v, "request": self.request}

        values = v.editbasics()
        values.update(vrender)
        render("nive.components.adminview:form.pt", values)
        
        values = v.tools()
        values.update(vrender)
        render("nive.components.adminview:tools.pt", values)
        
        render("nive.components.adminview:modules.pt", vrender)
        render("nive.components.adminview:root.pt", vrender)
        render("nive.components.adminview:views.pt", vrender)
        render("nive.components.adminview:help.pt", values)

    




# Copyright 2012, 2013 Arndt Droullier, Nive GmbH. All rights reserved.
# Released under GPL3. See license.txt
#

__doc__ = """
User and security functions.
"""

# imports used by other files
from pyramid.security import Allow 
from pyramid.security import Deny
from pyramid.security import ALL_PERMISSIONS
from pyramid.security import Everyone, Authenticated
from pyramid.security import remember, forget, authenticated_userid

from pyramid import threadlocal
from pyramid.interfaces import IAuthenticationPolicy, IAuthorizationPolicy

from nive.definitions import Conf
from nive.definitions import Interface, implementer


def GetUsers(app):
    """
    Loads all users from user database if available.
    """
    portal = app.portal
    try:
        userdb = portal.userdb
        return userdb.root.GetUsers()
    except:
        return []

    
class User(object):
    """
    A fake user object for testing.
    """
    def __str__(self):
        return self.data.name
    
    def __init__(self, name, id=0):
        self.id = id
        self.groups = []
        self.meta = Conf(title=name,pool_state=1)
        self.data = Conf(name=name,email="",groups=[])

    @property
    def identity(self):
        return str(self)
        
    def GetGroups(self, context=None):
        return self.groups
    
    def InGroups(self, groups):
        """
        check if user has one of these groups
        """
        if isinstance(groups, str):
            return groups in self.groups
        for g in groups:
            if g in self.groups:
                return True
        return False
    
    def ReadableName(self):
        return self.data.name
    
    def Commit(self, user=None):
        return


class Unauthorized(Exception):
    """
    failed login
    """

class UserFound(Exception):
    """
    Can be used in *getuser* listeners to break user lookup and
    pass a session user to LookupUser. The second argument is the session
    user 
    """
    def __init__(self, user):
        self.user = user



class IAdminUser(Interface):
    """
    Used for admin user instance hard coded in configration 
    """


@implementer(IAdminUser)
class AdminUser(object):
    """
    Admin User object with groups and login possibility. 
    """

    def __init__(self, values, ident):
        self.id = 0
        self.data = Conf(**values)
        self.meta = Conf()
        self.identity = ident or str(self.id)
        if values.get("groups"):
            groups = tuple(values.get("groups"))
        else:
            groups = ("group:admin",)
        self.groups = self.data.groups = groups

    def __str__(self):
        return str(self.identity)

    def Authenticate(self, password):
        return password == str(self.data["password"])
    
    def Login(self):
        """ """

    def Logout(self):
        """ """

    def GetGroups(self, context=None):
        """ """
        return self.groups

    def InGroups(self, groups):
        """
        check if user has one of these groups
        """
        if isinstance(groups, str):
            return groups in self.groups
        for g in groups:
            if g in self.groups:
                return True
        return False
    
    def ReadableName(self):
        return self.data.name


def SetupRuntimeAcls(acl, context):
    """
    Can be used during object initialisation to process acls.
    Acls can have a additional callback parameter called with the context
    to find out whetther to apply the single acl or not. The callback should return
    true or false. e.g. ::

        (Allow, "group:reader", "read", lambda context: context.meta.pool_state),

    or ::

        def check(context):
            return context.meta.pool_state

        (Allow, "group:reader", "read", check),

    :param acl:
    :param context:
    :return: tuple of acls to be stored as __acl__
    """
    processed = []
    for a in acl:
        if len(a)==4:
            if a[3](*(context,)):
                processed.append(tuple(list(a[:3])))
            continue
        processed.append(tuple(a))
    return tuple(processed)


def effective_principals(request=None):
    # uses pyramid authentication groupfinder callback to lookup principals
    # returns None if no auth policy is configured (e.g. in tests)
    registry = threadlocal.get_current_registry()
    request = request or threadlocal.get_current_request()
    authn_policy = registry.queryUtility(IAuthenticationPolicy)
    authz_policy = registry.queryUtility(IAuthorizationPolicy)
    if authn_policy and authz_policy:
        principals = authn_policy.effective_principals(request)
        return principals
    return None


        


# Copyright 2012, 2013 Arndt Droullier, Nive GmbH. All rights reserved.
# Released under GPL3. See license.txt
#

__doc__ = """
Data Pool Postgres Module
----------------------
*Requires psycopg2*. Use 'pip install python-psycopg2' to install the package.
"""


import threading

from nive.definitions import ConfigurationError
try:
    import psycopg2
    psycopg2.extensions.register_type(psycopg2.extensions.UNICODE)
    DEC2FLOAT = psycopg2.extensions.new_type(
        psycopg2.extensions.DECIMAL.values,
        'DEC2FLOAT',
        lambda value, curs: float(value) if value is not None else None)
    psycopg2.extensions.register_type(DEC2FLOAT)
except ImportError:
    # define fake psycopg class here to avoid test import errors if psycopg is not installed
    class psycopg2(object):
        OperationalError = None
        ProgrammingError = None
        Warning = None
        @staticmethod
        def connect(*args,**kw):
            raise ImportError("Python postgres binding not available. Try 'pip install psycopg2' to install the package.")
            


from nive.utils.utils import STACKF

from nive.utils.dataPool2.base import Base, Entry
from nive.utils.dataPool2.base import NotFound
from nive.utils.dataPool2.connection import Connection, ConnectionThreadLocal, ConnectionRequest

from nive.utils.dataPool2.postgres.dbManager import PostgresManager
from nive.utils.dataPool2.files import FileManager, FileEntry

from nive.definitions import OperationalError



class PostgresConnection(Connection):
    """
    postgres connection handling class

    config parameter in dictionary:
    dbName = initial database name
    user = database user

    password = password user
    host = host postgres server
    port = port postgres server

    user/host/port/password not mandatory if unix socket connection
    """
    #TODO add all psycopg2 connect options

    def connect(self):
        """ Close and connect to server """

        db = self.PrivateConnection()
        if not db:
            conf = self.configuration
            raise OperationalError("Cannot connect to database '%s.%s'" % (conf.host, conf.dbName))
        self._set(db)
        return db

    def IsConnected(self):
        """ Check if database is connected """
        db = self._get()
        return db and not db.closed

    def GetDBManager(self):
        """ returns the database manager obj """
        self.VerifyConnection()
        manager = PostgresManager()
        manager.SetDB(self.PrivateConnection())
        return manager

    def PrivateConnection(self):
        """ This function will return a new and raw connection. It is up to the caller to close this connection. """
        conf = self.configuration
        kargs = {'database': conf.dbName,
                'connect_timeout': conf.timeout}

        if conf.host:
            kargs['host'] = conf.host
        if conf.port:
            kargs['port'] = conf.port
        if conf.user:
            kargs['user'] = conf.user

        return psycopg2.connect(**kargs)

    def FmtParam(self, param):
        """ 
        Format a parameter for sql queries like literal for db. This function is not
        secure for any values. 
        """
        if isinstance(param, (int, float)):
            return str(param)
        d = str(param)
        if d.find("'")!=-1:
            d = d.replace("'","\\'")
        return "'%s'"%d


class PgConnThreadLocal(PostgresConnection, ConnectionThreadLocal):
    """
    Caches database connections as thread local values.
    """

    def __init__(self, config = None, connectNow = False):
        PostgresConnection.__init__(self, config, connectNow)


class PgConnRequest(PostgresConnection, ConnectionRequest):
    """
    Caches database connections as request values. Uses thread local stack as fallback (e.g testing).
    """

    def __init__(self, config = None, connectNow = False):
        self.local = threading.local()
        PostgresConnection.__init__(self, config, connectNow)

class PostgreSql(FileManager, Base):
    """
    Data Pool Postgres implementation
    """
    _OperationalError = psycopg2.OperationalError
    _ProgrammingError = psycopg2.ProgrammingError
    _Warning = psycopg2.Warning
    _DefaultConnection = PgConnRequest#PgConnThreadLocal
            

    def GetContainedIDs(self, base=0, sort="title", parameter=""):
        """
        select list of all entries
        id needs to be first field, pool_unitref second
        """
        def _SelectIDs(base, ids, sql, cursor):
            cursor.execute(sql%(base))
            entries = cursor.fetchall()
            for e in entries:
                ids.append(e[0])
                ids = _SelectIDs(e[0], ids, sql, cursor)
            return ids
        if parameter != "":
            parameter = "AND " + parameter
        parameter = ("WHERE pool_unitref=%d", parameter)
        sql = """SELECT id FROM %s %s ORDER BY %s""" % (self.MetaTable, " ".join(parameter), sort)
        cursor = self.connection.cursor()
        ids = _SelectIDs(base, [], sql, cursor)
        cursor.close()
        return ids
    
    
    def _GetInsertIDValue(self, cursor):
        cursor.execute("SELECT LASTVAL()")
        return cursor.fetchone()[0]

    def _CreateNewID(self, table = "", dataTbl = None):
        aC = self.connection.cursor()
        if table == "":
            table = self.MetaTable
        if table == self.MetaTable:
            if not dataTbl:
                raise ConfigurationError("Missing data table - Entry not created")

            # sql insert empty rec in meta table
            if self._debug:
                STACKF(0,"INSERT INTO %s DEFAULT VALUES" % (dataTbl)+"\r\n",self._debug, self._log,name=self.name)
            try:
                aC.execute("INSERT INTO %s DEFAULT VALUES" % (dataTbl))
            except self._Warning:
                pass
            # sql get id of created rec
            if self._debug:
                STACKF(0,"SELECT LASTVAL()\r\n",0, self._log,name=self.name)
            aC.execute("SELECT LASTVAL()")
            dataref = aC.fetchone()[0]
            # sql insert empty rec in meta table
            if self._debug:
                STACKF(0,"INSERT INTO %s (pool_datatbl, pool_dataref) VALUES ('%s', %s)"% (self.MetaTable, dataTbl, dataref),0, self._log,name=self.name)
            aC.execute("INSERT INTO %s (pool_datatbl, pool_dataref) VALUES ('%s', %s)"% (self.MetaTable, dataTbl, dataref))
            if self._debug:
                STACKF(0,"SELECT LAST_INSERT_ID()\r\n",0, self._log,name=self.name)
            aC.execute("SELECT LASTVAL()")
            aID = aC.fetchone()[0]
            aC.close()
            return aID, dataref

        # sql insert empty rec in meta table
        if self._debug:
            STACKF(0,"INSERT INTO %s () VALUES ()" % (table)+"\r\n",self._debug, self._log,name=self.name)
        aC.execute("INSERT INTO %s () VALUES ()" % (table))
        # sql get id of created rec
        if self._debug:
            STACKF(0,"SELECT LAST_INSERT_ID()\r\n",0, self._log,name=self.name)
        aC.execute("SELECT LAST_INSERT_ID()")
        aID = aC.fetchone()[0]
        aC.close()
        return aID, 0

    # types/classes -------------------------------------------------------------------


    def _GetPoolEntry(self, id, **kw):
        try:
            return PostgreSqlEntry(self, id, **kw)
        except NotFound:
            return None


    def _FmtLimit(self, start, max):
        if start != None:
            return "LIMIT %s OFFSET %s" % (str(max), str(start))
        return "LIMIT %s" % (str(max))


class PostgreSqlEntry(FileEntry, Entry):
    """
    Data Pool Entry Postgres implementation
    """



# Copyright 2012, 2013 Arndt Droullier, Nive GmbH. All rights reserved.
# Released under GPL3. See license.txt
#

import json
from datetime import datetime, timedelta
from datetime import time as datetime_time
from decimal import Decimal

from nive.utils.utils import ConvertToDateTime
from nive.utils.dataPool2.files import File


# Database records wrapper ---------------------------------------------------------------------------

class Wrapper(object):
    """
    Wrappers are mapping objects for data, files and meta. Content can be accessed as
    dictionary field.
    Changes are stored temporarily in memory.
    """

    __wrapper__ = 1
    meta = False
    
    def __init__(self, entry, content=None):
        self._entry_ = entry
        self._temp_ = {}
        self._content_ = None
        
    def __repr__(self):
        return str(type(self)) 
    
    def __dir__(self):
        return ["_temp_", "_content_", "_entry_"]
    
    def __setitem__(self, key, value):
        if key in ("id", "pool_datatbl", "pool_dataref"):
            return
        self._temp_[key] = self._entry_().DeserializeValue(key, value, self.meta)

    def __getitem__(self, key):
        if key in self._temp_:
            return self._temp_[key]
        if not self._content_:
            self._Load()
        return self._content_.get(key)

    def __contains__(self, key):
        if key in self._temp_:
            return True
        if not self._content_:
            self._Load()
        return key in self._content_

    def __getattr__(self, key):
        if key in list(self.__dict__.keys()):
            return self.__dict__[key]
        if key in self._temp_:
            return self._temp_[key]
        if not self._content_:
            self._Load()
        return self._content_.get(key)


    def close(self):
        self._entry_ = None
        self._temp_.clear()
        self._content_ = None


    def clear(self):
        """
        Reset contents, temp data and entry obj
        """
        self._temp_.clear()


    def copy(self):
        """
        Returns a copy of current content
        """
        if not self._content_:
            self._Load()
        c = self._content_.copy()
        c.update(self._temp_)
        return c


    def has_key(self, key):
        if self.HasTempKey(key):
            return True
        return key in list(self.keys())


    def get(self, key, default=None):
        try:
            data = self[key]
            if data is None:
                return default
            return data
        except AttributeError:
            return default


    def set(self, key, data):
        self[key] = data


    def update(self, dict, force = False):
        dict = self._entry_().DeserializeValue(None, dict, self.meta)
        if force:
            for k in list(dict.keys()):
                data = dict[k]
                if isinstance(data, bytes):
                    dict[k] = self._entry_().pool.DecodeText(data)
            self._temp_.update(dict)
            return
        for k in list(dict.keys()):
            self[k] = dict[k]


    def keys(self):
        if not self._content_:
            self._Load()
        t = list(self._content_.keys())
        t += list(self._temp_.keys())
        return t



    def IsEmpty(self):                return self._content_ is None
    def GetTemp(self):                return self._temp_
    def HasTemp(self):                return self._temp_ != {}
    def GetTempKey(self, key):        return self._temp_.get(key)
    def HasTempKey(self, key):        return key in self._temp_

    def GetEntry(self):                return self._entry_()

    def SetContent(self, content):
        if not self._content_:
            self._content_ = content
        else:
            self._content_.update(content)

    def EmptyTemp(self):
        self._temp_.clear()

    def _Load(self):
        self._content_ = {}
        pass


class MetaWrapper(Wrapper):
    """
    wrapper class for meta content
    """
    meta = True
    
    def _Load(self):
        self._content_ = {}
        self._entry_()._PreloadMeta()



class DataWrapper(Wrapper):
    """
    wrapper class for data content
    """

    def _Load(self):
        self._content_ = {}
        self._entry_()._PreloadData()


class FileWrapper(Wrapper):
    """
    wrapperclass for files. contains only filemta and returns file streams on read.
    update and __setitem__ take File object with o.file and o.filename attr as parameter
    entry = {"filename": "", "path": <absolute path for temp files>, "file": <file stream>}
    """

    def __setitem__(self, key, filedata):
        """
        filedata can be a dictionary, File object or file path
        """
        if not filedata:
            if key in self._temp_:
                del self._temp_[key]
            elif self._content_ and key in self._content_:
                del self._content_[key]
            return
        if isinstance(filedata, dict):
            file = File(key, filedict=filedata, fileentry=self._entry_())
            filedata = file
        elif isinstance(filedata, bytes):
            # load from temp path
            file = File(key, fileentry=self._entry_())
            file.fromPath(filedata)
            filedata = file
        filedata.tempfile = True
        self._temp_[key] = filedata


    def set(self, key, filedata):
        self[key] = filedata


    def SetContent(self, files):
        self._content_ = {}
        if isinstance(files, dict):
            for f in files:
                self._content_[f] = files[f]
            return
        for f in files:
            self._content_[f["filekey"]] = f


    def _Load(self):
        files = self._entry_().Files()
        self._content_ = {}
        for f in files:
            self._content_[f["filekey"]] = f
        return list(self._content_.keys())


#  Pool Structure ---------------------------------------------------------------------------


class PoolStructure(object):
    """
    Data Pool 2 Structure handling. Defines a table field mapping. If field types are available serializing 
    and deserializing is performed on database reads and writes.

    ::
    
        structure =
            {
             meta:   (field1, field2, ...),
             type1_table: (field5, field6, ...),
             type2_table: (field8, field9, ...),
            }

        fieldtypes = 
            {
             meta: {field1: string, field2: number},
             type1_table: {field5: DateTime, field6: text},
             type2_table: {field8: DateTime, field9: text},
            }
            
        stdMeta = (field1, field2)

    Deserialization datatypes ::
    
        string, htext, text, list, code, radio, email, password, url -> unicode
        number, float, unit -> number 
        bool -> 0/1
        file -> bytes
        timestamp -> float
        decimal -> float
        date, datetime -> datetime
        multilist, checkbox, urllist -> unicode tuple
        unitlist -> number tuple
        json -> python type list, tuple or dict

    Serialization datatypes ::
    
        string, htext, text, list, code, radio, email, password, url -> unicode
        number, float, unit -> number 
        bool -> 0/1
        file -> bytes
        timestamp -> float
        date, datetime -> datetime
        multilist, checkbox, urllist -> json
        unitlist -> json
        json -> json
        
    If fieldtype (`fieldtypes`) information is not given json data is stored with `_json_`
    prefix.

    """
    MetaTable = "pool_meta"
    
    def __init__(self, structure=None, fieldtypes=None, stdMeta=None, codepage="utf-8", tzinfo=None, **kw):
        #
        self.stdMeta = ()
        self.structure = {}
        self.fieldtypes = {}
        self.serializeCallbacks = {}
        self.deserializeCallbacks = {}
        self.codepage = codepage
        self.pytimezone = tzinfo
        if structure:
            self.Init(structure, fieldtypes, stdMeta, codepage, tzinfo, **kw)
        

    def Init(self, structure, fieldtypes=None, stdMeta=None, codepage="utf-8", tzinfo=None, **kw):
        s = structure.copy()
        self.codepage = codepage
        self.pytimezone = tzinfo
        meta = list(s[self.MetaTable])
        # add default fields
        if not "pool_dataref" in s[self.MetaTable]:
            meta.append("pool_dataref")
        if not "pool_datatbl" in s[self.MetaTable]:
            meta.append("pool_datatbl")
        s[self.MetaTable] = tuple(meta)
        for k in s:
            s[k] = tuple(s[k])
        self.structure = s
        
        if fieldtypes:
            self.fieldtypes = fieldtypes
        
        if stdMeta:
            self.stdMeta = tuple(stdMeta)


    def IsEmpty(self):
        return self.structure=={}

    
    def __getitem__(self, key, version=None):
        return self.structure[key]

    def __contains__(self, key):
        return key in self.structure

    def get(self, key, default=None, version=None):
        return self.structure.get(key, default)

    def has_key(self, key, version=None):
        return key in self.structure

    def keys(self, version=None):
        return list(self.structure.keys())
    
    
    def serialize(self, table, field, value):
        # if field==None and value is a dictionary multiple values are serialized
        if field is None and isinstance(value, dict):
            newdict = {}
            for field, v in list(value.items()):
                try:        t = self.fieldtypes[table][field]
                except:     t = None
                newdict[field] = self._se(v, t, field)
            return newdict
        else:
            try:        t = self.fieldtypes[table][field]
            except:     t = None
            value = self._se(value, t, field)
        return value
        

    def deserialize(self, table, field, value):
        # if field==None and value is a dictionary multiple values are deserialized
        if field is None and isinstance(value, dict):
            newdict = {}
            for field, v in list(value.items()):
                try:        t = self.fieldtypes[table][field]
                except:     t = None
                newdict[field] = self._de(v, t, field)
            return newdict
        else:
            try:        t = self.fieldtypes[table][field]
            except:     t = None
            value = self._de(value, t, field)
        return value


    def _se(self, value, fieldtype, field):
        if not fieldtype:
            # no datatype information set
            if isinstance(value, datetime):
                return value.strftime("%Y-%m-%d %H:%M:%S")
            elif isinstance(value, timedelta):  # treat as time. py3 mysql bug?
                n = datetime(year=2000, month=1, day=1, hour=0, minute=0, second=0) + value
                return n.strftime("%H:%M:%S")
            elif isinstance(value, (list, tuple)):
                if isinstance(value[0], bytes):
                    # list of strings:
                    value = [str(v, self.codepage) for v in value]
                value = "_json_"+json.dumps(value)
            elif isinstance(value, bytes):
                value = str(value, self.codepage)
            return value
        
        if isinstance(fieldtype, dict):
            fieldtype = fieldtype["datatype"]
            
        # call serialize callback function
        if fieldtype in self.serializeCallbacks:
            return self.serializeCallbacks[fieldtype](value, field)
            
        if fieldtype == "number":
            if isinstance(value, str):
                value = int(value)
            elif isinstance(value, float):
                value = int(value)
        
        elif fieldtype == "float":
            if isinstance(value, str):
                value = float(value)

        elif fieldtype in ("date", "datetime"):
            if isinstance(value, (float,int)):
                value = str(datetime.fromtimestamp(value))
            if isinstance(value, datetime):
                value = value.astimezone(self.pytimezone).strftime("%Y-%m-%d %H:%M:%S")
            elif not isinstance(value, str) and not value is None:
                value = str(value)
        
        elif fieldtype == "time":
            if isinstance(value, (float,int)):
                value = str(datetime.fromtimestamp(value).strftime("HH:MM:SS.%f"))
            elif isinstance(value, timedelta):  # treat as time. py3 mysql bug?
                n = datetime(year=2000, month=1, day=1, hour=0, minute=0, second=0) + value
                value = n.strftime("HH:MM:SS.%f")
            elif value is None:
                pass
            elif not isinstance(value, str):
                value = str(value)

        elif fieldtype == "timestamp":
            if value is None:
                pass
            elif isinstance(value, Decimal):
                value = float(value)
        
        elif fieldtype in ("list","radio"):
            # to single item string
            if isinstance(value, (list, tuple)):
                if value:
                    value = value[0]
                else:
                    value = ""

        elif fieldtype in ("multilist", "checkbox", "mselection", "mcheckboxes", "urllist", "unitlist"):
            # to json formatted list
            if not value:
                value = ""
            elif isinstance(value, str):
                value = [value]
            if isinstance(value, (list, tuple)):
                if isinstance(value[0], bytes):
                    # list of strings:
                    value = [str(v, self.codepage) for v in value]
                value = json.dumps(value)

        elif fieldtype in ("bool",):
            if isinstance(value, str):
                if value.lower()=="true":
                    value = 1
                elif value.lower()=="false":
                    value = 0
            else:
                try:
                    value = int(value)
                except:
                    value = 0

        elif fieldtype == "json":
            if not value:
                value = ""
            elif not isinstance(value, str):
                value = json.dumps(value)
            
        # assure unicode except filedata
        if isinstance(value, bytes) and fieldtype!="file":
            value = str(value, self.codepage)

        return value


    def _de(self, value, fieldtype, field):
        if not fieldtype:
            # no datatype information set
            if isinstance(value, str) and value.startswith("_json_"):
                value = json.loads(value[len("_json_"):])
            elif isinstance(value, timedelta):  # treat as time. py3 mysql bug?
                n = datetime(year=2000, month=1, day=1, hour=0, minute=0, second=0) + value
                value = datetime_time(hour=n.hour, minute=n.minute, second=n.second, microsecond=n.microsecond)
            if isinstance(value, bytes):
                value = str(value, self.codepage)
            return value

        if isinstance(fieldtype, dict):
            fieldtype = fieldtype["datatype"]

        # call serialize callback function
        if fieldtype in self.deserializeCallbacks:
            return self.deserializeCallbacks[fieldtype](value, field)

        if fieldtype in ("date", "datetime"):
            # -> to datetime
            if isinstance(value, str):
                value = ConvertToDateTime(value)
            elif isinstance(value, (float,int)):
                value = datetime.fromtimestamp(value)
            if value is not None and self.pytimezone is not None and value.tzinfo is None:
                value = value.replace(tzinfo=self.pytimezone)
                    
        elif fieldtype == "time":
            # -> to datetime.time
            if isinstance(value, str):
                # misuse datetime parser
                value2 = ConvertToDateTime("2015-01-01 "+str(value))
                if value2:
                    value = datetime_time(value2.hour,value2.minute,value2.second,value2.microsecond)
            elif isinstance(value, (float,int)):
                value = datetime.fromtimestamp(value)
                value = datetime_time(value.hour,value.minute,value.second,value.microsecond)
            elif isinstance(value, timedelta):  # treat as time. py3 mysql bug?
                n = datetime(year=2000, month=1, day=1, hour=0, minute=0, second=0) + value
                value = datetime_time(hour=n.hour, minute=n.minute, second=n.second, microsecond=n.microsecond)

        elif fieldtype == "timestamp":
            if isinstance(value, str):
                value = float(value)
                    
        elif fieldtype in ("multilist", "checkbox", "mselection", "mcheckboxes", "urllist"):
            # -> to string tuple
            # unitlist -> to number tuple
            if not value:
                value = ""
            elif isinstance(value, str):
                if value.startswith("_json_"):
                    value = json.loads(value[len("_json_"):])
                else:
                    try:
                        value = tuple(json.loads(value))
                    except ValueError:
                        # use as single item text string
                        value = (value,)
            elif isinstance(value, list):
                value = tuple(value)
            if fieldtype == "unitlist":
                value = tuple([int(v) for v in value])

        elif fieldtype in ("unitlist",):
            # -> to string tuple
            # unitlist -> to number tuple
            if not value:
                value = ""
            elif isinstance(value, str):
                if value.startswith("_json_"):
                    value = json.loads(value[len("_json_"):])
                else:
                    try:
                        value = tuple(json.loads(value))
                    except ValueError:
                        # use as single item text string
                        value = (value,)
                    except TypeError:
                        # use as single item text string
                        value = (value,)
            elif isinstance(value, list):
                value = [int(v) for v in value]
                value = tuple(value)

        elif fieldtype == "json":
            # -> to python type
            if not value:
                value = None
            elif isinstance(value, str):
                value = json.loads(value)

        return value
    
    


# Copyright 2012, 2013 Arndt Droullier, Nive GmbH. All rights reserved.
# Released under GPL3. See license.txt
#

import weakref
import json
import os
import shutil
import hashlib
from datetime import datetime, timedelta
from datetime import time as datetime_time

from pyramid.path import DottedNameResolver
from pyramid.path import AssetResolver
from pyramid.path import caller_package

from pyramid.renderers import JSON

from nive.definitions import (
    IAppConf, IDatabaseConf, IFieldConf, IRootConf, IObjectConf, IViewModuleConf,
    IViewConf, IToolConf, IPortalConf, IGroupConf, IModuleConf,
    IWidgetConf, IWfProcessConf, IWfStateConf, IWfTransitionConf, IConf, IFileStorage
)
from nive.definitions import Conf
from nive.definitions import baseConf
from nive.definitions import ConfigurationError
from nive import File
from nive.i18n import _



def ResolveName(name, base=None, raiseExcp=True):
    """
    Lookup python object by dotted python name.
    Wraps pyramid.DottedNameResolver.
    
    returns object or None
    """
    if not name:
        return None
    if not isinstance(name, str):
        return name
    if not base:
        base = caller_package()
    try:
        return DottedNameResolver(base).resolve(name)
    except:
        if not raiseExcp:
            return None
        raise

def ResolveAsset(name, base=None, raiseExcp=True):
    """
    Lookup asset path (template, json or any other file) and returns asset 
    descriptor object or None.
    """
    if not name:
        return None
    if not isinstance(name, str):
        return name
    if not base:
        base = caller_package()
    if name.startswith("./"):
        # use relative file system path
        name = os.getcwd()+name[1:]
    try:
        return AssetResolver(base).resolve(name)
    except:
        if not raiseExcp:
            return None
        raise

    
def ResolveConfiguration(conf, base=None):
    """
    Lookup configuration object by dotted python name. Returns interface and configuration object.
    Extends pyramid.DottedNameResolver with .json file support for configuration objects.
    
    Supports the following cases:
    
    - Path and file name to .json file. requires `type` set to one of the 
      configuration types: 
      *AppConf, FieldConf, DatabaseConf, RootConf, ObjectConf, ViewModuleConf, ViewConf, ToolConf, 
      GroupConf*
    - Dotted python name for configuration object including attribute name of configuration instance.
    - Dotted python name for object. Uses the convention to load the configuration from the 
      'configuration' attribute of the referenced object.
    - Configuration instance. Will just return it.
    
    returns Interface, configuration
    """
    # string instance
    if isinstance(conf, str):
        if not base:
            base = caller_package()
        # json file
        if conf.find(".json")!= -1:
            path = ResolveAsset(conf, base)
            with open(path.abspath()) as f:
                s = f.read()
            conf = json.loads(s)
        # resolve attribute name
        elif conf:
            conf = ResolveName(conf, base=base)

    # dict instance
    elif isinstance(conf, dict):
        # load by interface
        if not "type" in conf:
            raise TypeError("Configuration type not defined")
        c = ResolveName(conf["type"], base="nive")
        del conf["type"]
        conf = c(**conf)

    # module and not configuration
    if not isinstance(conf, baseConf):
        if hasattr(conf, "configuration"):
            conf = conf.configuration
        
    # object instance
    if IAppConf.providedBy(conf): return IAppConf, conf
    if IDatabaseConf.providedBy(conf): return IDatabaseConf, conf
    if IFieldConf.providedBy(conf): return IFieldConf, conf
    if IRootConf.providedBy(conf): return IRootConf, conf
    if IObjectConf.providedBy(conf): return IObjectConf, conf
    if IViewModuleConf.providedBy(conf): return IViewModuleConf, conf
    if IViewConf.providedBy(conf): return IViewConf, conf
    if IToolConf.providedBy(conf): return IToolConf, conf
    if IPortalConf.providedBy(conf): return IPortalConf, conf
    if IGroupConf.providedBy(conf): return IGroupConf, conf
    if IModuleConf.providedBy(conf): return IModuleConf, conf
    if IWidgetConf.providedBy(conf): return IWidgetConf, conf
    if IWfProcessConf.providedBy(conf): return IWfProcessConf, conf
    if IWfStateConf.providedBy(conf): return IWfStateConf, conf
    if IWfTransitionConf.providedBy(conf): return IWfTransitionConf, conf
    if IConf.providedBy(conf): return IConf, conf
    return None, conf
    #raise TypeError, "Unknown configuration object: %s" % (str(conf))
    
    
def LoadConfiguration(conf, base=None):
    """
    same as ResolveConfiguration except only the configuration object is
    returned
    """
    if not base:
        base = caller_package()
    i,c = ResolveConfiguration(conf, base)
    return c


def FormatConfTestFailure(report, fmt="text"):
    """
    Format configuration test() failure
    
    returns string
    """
    import inspect
    v=[]
    for r in report:
        v+= "-----------------------------------------------------------------------------------\r\n"
        v+= str(r[0]) + " " + r[1] + "\r\n"
        v+= "-----------------------------------------------------------------------------------\r\n"
        for d in list(r[2].__dict__.items()):
            a = d[1]
            if a is None:
                try:
                    a = r[2].parent.get(d[0])
                except:
                    pass
            v+= str(d[0])+":  "+str(a)+"\r\n"
        v+= "\r\n"
    return "".join(v)


def ReplaceInListByID(conflist, newconf, id=None):
    new = []
    id = id or newconf.id
    for conf in conflist:
        if conf.id == id:
            new.append(newconf)
        else:
            new.append(conf)
    return new
            
def UpdateInListByID(conflist, updateconf, id=None):
    id = id or updateconf.id
    for conf in conflist:
        if conf.id == id:
            conf.update(updateconf)
    return conflist


class JsonDataEncoder(json.JSONEncoder):
    exportDirectory = None  # set a valid path name to activate file exports. file["path"] points to the exported file
    hashFilename = False
    def default(self, obj):
        if isinstance(obj, datetime):
            return str(obj)
        elif isinstance(obj, datetime_time):
            return str(obj)
        elif isinstance(obj, timedelta):  # treat as time. py3 mysql bug?
            n = datetime(year=2000, month=1, day=1, hour=0, minute=0, second=0) + obj
            return n.strftime("HH:MM:SS.%f")
        elif IFileStorage.providedBy(obj):
            file = dict(filekey = obj.filekey, filename = obj.filename, size = obj.size)
            fn = "%s-%d-%s" % (obj.uid, obj.fileid, obj.filename)
            if self.hashFilename:
                fn = hashlib.md5(fn.encode("utf-8")).hexdigest() + "." + obj.extension
                file["path"] = fn
            if self.exportDirectory and obj.path:
                # copy file
                dest = self.exportDirectory + fn
                src = obj.abspath()
                shutil.copyfile(src, dest)
                file["path"] = fn
            return file
        return json.JSONEncoder.default(self, obj)

class ConfEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, baseConf):
            values = {}
            for k in obj:
                values[k] = obj[k]
            return values
        elif isinstance(obj, datetime):
            return str(obj)
        elif isinstance(obj, datetime_time):
            return str(obj)
        elif isinstance(obj, timedelta):
            return obj.isoformat()
        return json.JSONEncoder.default(self, obj)

class ConfDecoder(object):
    def decode(self, jsonstring):
        def object_hook(obj):
            if isinstance(obj, dict):
                try:
                    confclass = obj["ccc"]
                except KeyError:
                    return obj
                   
                if not confclass:
                    raise ConfigurationError("Configuration class not found (ccc)")
                conf = ResolveName(confclass, base="nive")(**obj)
                return conf
            return obj
        return json.JSONDecoder(object_hook=object_hook).decode(jsonstring) 


def SetupJSONRenderer(pyramid_config):
    """
    Extend the default pyramid json renderer with support for datetime and file storage
    objects. Call `SetupJSONRenderer` in the main function e.g. ::

        config = Configurator(root_factory = getRoot, settings = settings)
        SetupJSONRenderer(config)
        config.include('pyramid_chameleon')

    :param pyramid_config:
    :return:
    """
    json_renderer = JSON()

    def datetimeAdapter(obj, request):
        return obj.isoformat()
    json_renderer.add_adapter(datetime, datetimeAdapter)

    def fileStorageAdapter(obj, request):
        file = {}
        file["filekey"] = obj.filekey
        file["filename"] = obj.filename
        file["size"] = obj.size
        return file
    json_renderer.add_adapter(IFileStorage, fileStorageAdapter)

    def confAdapter(obj, request):
        return DumpJSONConf(obj)
    json_renderer.add_adapter(baseConf, confAdapter)

    pyramid_config.add_renderer('json', json_renderer)


def DumpJSONConf(conf):
    # dump configuration to json
    values = {}
    for k in conf:
        v = conf[k]
        if isinstance(v, baseConf):
            values[k] = DumpJSONConf(v)
        elif isinstance(v, datetime):
            values[k] = str(v)
        else:
            values[k] = v
    return json.dumps(values)

def LoadJSONConf(jsondata, default=None):
    # jsondata must be a json string or dictionary
    # load from json
    # default: the default configuration class to be used if the json values do not
    # specify the class as `ccc`
    if isinstance(jsondata, str):
        try:
            jsondata = json.loads(jsondata)
        except:
            return jsondata
    if not isinstance(jsondata, dict):
        return jsondata
    for k,v in list(jsondata.items()):
        jsondata[k] = LoadJSONConf(v, default=default)
        
    confclass = jsondata.get("ccc")
    if not confclass:
        if not default:
            raise ConfigurationError("Configuration class not found (ccc)")
        return default(**jsondata)
    conf = ResolveName(confclass, base="nive")(**jsondata)
    return conf



def GetVirtualObj(configuration, app):
    """
    This loads an object for a non existing database entry.
    """
    if not configuration:
        raise ConfigurationError("Type not found")
    obj = ClassFactory(configuration, app.reloadExtensions, True, base=None)
    dbEntry = app.db.GetEntry(0, virtual=1)
    obj = obj(0, dbEntry, parent=None, configuration=configuration)
    return obj


def ClassFactory(configuration, reloadClass=False, raiseError=True, base=None, storeConfAsStaticClassVar=False):
    """
    Creates a python class reference from configuration. Uses configuration.context as class
    and dynamically adds classes listed as configuration.extensions as base classes.
    
    configuration requires

    - configuration.context
    - configuration.extensions [optional]
    
    If reloadClass = False the class is cached as configuration._c_class.
    
    storeConfAsStaticClassVar: experimental option. If true stores the configuration as part of the 
    class.
    
    """
    if not reloadClass:
        try:
            return configuration._c_class
        except AttributeError:
            pass
    tag = configuration.context
    if "extensions" in configuration:
        bases = configuration.extensions
    else:
        bases = None
    cls = GetClassRef(tag, reloadClass, raiseError, base)
    if not cls:
        return None
    
    def cacheCls(configuration, cls):
        # store type() class
        lock = configuration.locked
        if lock:
            configuration.unlock()
        configuration._c_class = cls
        if lock:
            configuration.lock()
    
    if not bases:
        if storeConfAsStaticClassVar:
            cls = type("_factory_"+cls.__name__, (cls,), {})
            cls.configuration = configuration
        cacheCls(configuration, cls)
        return cls

    # load extensions
    b = [cls]
    #opt
    for r in bases:
        r = GetClassRef(r, reloadClass, raiseError, base)
        if not r:
            continue
        b.insert(0,r)
    if len(b)==1:
        return cls

    # create new class with name configuration.context
    cls = type("_factory_"+cls.__name__, tuple(b), {})
    if storeConfAsStaticClassVar:
        cls.configuration = configuration
    cacheCls(configuration, cls)
    return cls


def GetClassRef(tag, reloadClass=False, raiseError=True, base=None):
    """
    Resolve class reference from python dotted string.
    """
    if isinstance(tag, str):
        if raiseError:
            classRef = ResolveName(tag, base=base)
        else:
            try:
                classRef = ResolveName(tag, base=base)
            except ImportError as e:
                return None
        if not classRef:
            return None
        #if reloadClass:
        #    reload(classRef)
        return classRef
    # tag is class ref
    return tag


def DecorateViewClassWithViewModuleConf(viewModuleConf, cls):
    ref = weakref.ref(viewModuleConf)
    if isinstance(cls, str):
        cls = ResolveName(cls)
    cls = type("_factory_"+cls.__name__, (cls,), {})
    cls.__configuration__ = ref
    return cls


# Field list items ------------------------------------------

from nive.security import GetUsers
from nive.utils.language import LanguageExtension, CountryExtension

def LoadListItems(fieldconf, app=None, obj=None, pool_type=None, force=False, user=None):
    """
    Load field list items for the given fieldconf.
    If `force` is False and fieldconf already contains list items, the existing 
    `fieldconf.listItems` are returned. Set `force=True` to reload each time this
    function is called.
        
    `obj` and `pool_type` are only used for workflow lookup.
        
    returns dict list
    """
    values = []
    if not fieldconf:
        return values

    if fieldconf.listItems and not force:
        # skip loading if list filled
        if hasattr(fieldconf.listItems, '__call__'):
            return fieldconf.listItems(field=fieldconf, context=obj or app, user=user)
        return fieldconf.listItems
    
    if app is None:
        if fieldconf.settings:
            # settings dyn list
            dyn = fieldconf.settings.get("codelist")
            if dyn in ("languages", "countries"):
                # the only two lists not requiring 'app'
                if dyn == "languages":
                    return LanguageExtension().Codelist()
                else:
                    return CountryExtension().Codelist()
            
        # abort here if app is not set
        return fieldconf.listItems

    # load list items from db, application or user database
    if fieldconf.settings:
        # settings dyn list
        dyn = fieldconf.settings.get("codelist") or ""
        obj_type = fieldconf.settings.get("obj_type")
        if dyn == "users":
            return GetUsers(app)
        elif dyn == "groups":
            portal = app.portal
            if portal is None:
                portal = app
            sort = fieldconf.settings.get("sort", "id")
            return portal.GetGroups(sort=sort, visibleOnly=True)
        elif dyn == "localgroups":
            return app.GetGroups(sort="id", visibleOnly=True)
        elif dyn == "groups+auth":
            portal = app.portal
            if portal is None:
                portal = app
            return [Conf(id="authenticated", name=_("Authenticated"), visible=True)] + portal.GetGroups(sort="id", visibleOnly=True)
        elif dyn == "types":
            return app.configurationQuery.GetAllObjectConfs()
        elif dyn == "categories":
            return app.configurationQuery.GetAllCategories()
        elif dyn[:5] == "type:":
            type = dyn[5:]
            return app.root.search.GetEntriesAsCodeList(type, "title", parameter= {}, operators = {}, sort = "title")
        elif dyn == "meta":
            return app.root.search.GetEntriesAsCodeList2("title", parameter= {}, operators = {}, sort = "title")
        elif dyn == "languages":
            return LanguageExtension().Codelist()
        elif dyn == "countries":
            return CountryExtension().Codelist()

        elif obj_type:
            ot = fieldconf.settings.get("obj_type")
            tf = fieldconf.settings.get("name_field", "title")
            operators = dict()
            parameter = dict()
            if ot:
                parameter["pool_type"] = ot
            if isinstance(ot, (list, tuple)):
                operators["pool_type"] = "IN"
            if fieldconf.settings.get("app"):
                app = app.portal[fieldconf.settings.get("app")]
            values = app.root.search.GetEntriesAsCodeList2(tf, parameter=parameter, operators=operators, sort=tf)
            if fieldconf.settings.get("add_title_id"):
                return [dict(id=str(a["id"]),name="%s (%d)"%(a["name"],a["id"])) for a in values]
            else:
                return [dict(id=str(a["id"]),name=a["name"]) for a in values]

    fld = fieldconf.id
    if fld == "pool_type":
        values = app.configurationQuery.GetAllObjectConfs(visibleOnly=True)

    elif fld == "pool_category":
        values = app.configurationQuery.GetAllCategories(visibleOnly=True)

    elif fld == "pool_groups":
        local = fieldconf.settings.get("local")
        loader = app
        if not local:
            portal = app.portal
            if portal:
                loader = portal
        values = loader.GetGroups(sort="id", visibleOnly=True)

    elif fld == "pool_language":
        values = app.GetLanguages()

    elif fld == "pool_wfa":
        # uses type object as param
        if obj:
            try:
                aWfp = obj.meta.get("pool_wfp")
                obj = app.GetWorkflow(aWfp)
                if obj:
                    values = obj.GetActivities()
            except:
                pass
        elif pool_type:
            aWfp = app.configurationQuery.GetObjectConf(pool_type).get("workflowID")
            try:
                obj = app.GetWorkflow(aWfp)
                values = obj.GetActivities()
            except:
                pass
        else:
            values = []

    elif fld == "pool_wfp":
        values = app.configurationQuery.GetAllWorkflowConfs()

    return values


# test request and response --------------------------------------

class Response(object):
    headerlist = []

class Request(object):
    POST = {}
    GET = {}
    url = ""
    username = ""
    response = Response()
    environ = {}
    def resource_url(self, o):
        return str(o)

class Event(object):
    def __init__(self, request=None):
        self.request = request or Request()

class FakeLocalizer(object):
    def translate(self, text, mapping=None):
        try:
            if text.mapping:
                v = str(text)
                for k in text.mapping:
                    v = v.replace("${%s}"%k, str(text.mapping[k]))
                return v
        except:
            pass
        return str(text)



#-*- coding: utf-8 -*-

from nive.utils.path import DvPath
from nive.definitions import *
from nive.portal import Portal
from nive.security import User, Allow, Everyone

from nive.application import Application
from nive.utils.dataPool2.files import File

root = RootConf(
    id = "root",
    name = "Data root",
    default = 1,
    context = "nive.container.Root",
    subtypes = [IObject],
    acl = ((Allow, Everyone, 'view'),)
)

type1 = ObjectConf(
    id = "type1",
    name = "Type 1 container",
    hidden = False,
    dbparam = "data1",
    context = "nive.container.Container",
    subtypes = [IObject],
    acl = ((Allow, Everyone, 'view'),)
)
type1.data = [
    FieldConf(id="ftext", datatype="text", size=1000, name="ftext", fulltext=1),
    FieldConf(id="fnumber", datatype="number", size=8, name="fnumber", required=1),
    FieldConf(id="fdate", datatype="datetime", size=8, name="fdate"),
    FieldConf(id="ftime", datatype="time",     size=8, name="ftime"),
    FieldConf(id="flist", datatype="list",     size=100, name="flist", listItems=[{"id": "item 1", "name":"Item 1"},{"id": "item 2", "name":"Item 2"},{"id": "item 3", "name":"Item 3"}]),
    FieldConf(id="fmselect", datatype="multilist", size=50, name="fmselect"),
    FieldConf(id="funit", datatype="unit", size=8, name="funit"),
    FieldConf(id="funitlist", datatype="unitlist", size=100, name="funitlist"),
    FieldConf(id="fjson", datatype="json", size=1000, name="fjson"),
]

type2 = ObjectConf(
    id = "type2",
    name = "Type 2 Object",
    hidden = False,
    dbparam = "data2",
    context = "nive.objects.Object",
    subtypes = None
)
data2 = []
data2.append(FieldConf(**{"id":"fstr", "datatype": "string", "size": 100, "name": "fstr", "fulltext": 1}))
data2.append(FieldConf(**{"id":"ftext", "datatype": "text", "size": 1000, "name": "ftext"}))
data2.append(FieldConf(**{"id":"file1", "datatype": "file", "size": 5*1024*1024, "name": "file1"}))
data2.append(FieldConf(**{"id":"file2", "datatype": "file", "size": 500*1024, "name": "file2"}))
type2.data = data2

def create_callback(context, **kw):
    context.__kws_test__ = kw
    pass

type3 = ObjectConf(
    id = "type3",
    name = "Type 3 Object",
    hidden = True,
    dbparam = "data3",
    context = "nive.container.Container",
    subtypes = [INonContainer],
    events = (Conf(event="create", callback=create_callback),)
)
data3 = []
data3.append(FieldConf(**{"id":"fstr", "datatype": "string", "size": 100, "name": "fstr", "fulltext": 1}))
type3.data = data3

cat1 = Conf(**{"id":"cat1", "name":"Category 1"})
cat2 = Conf(**{"id":"cat2", "name":"Category 2"})

group1 = GroupConf(**{"id":"group1", "name":"Group 1"})
group2 = GroupConf(**{"id":"group2", "name":"Group 2"})

# configuration
appconf = AppConf(
    id = "unittest",
    context="nive.application.Application",
    title = "nive application python unittest",
    timezone = "UTC",
    modules = [root, type1, type2, type3],
    categories = [cat1, cat2],
    groups = [group1, group2],
    fulltextIndex = True,
    # run tests with the minimum required set of meta fields
    meta = copy.deepcopy(list(SystemFlds)) + copy.deepcopy(list(UserFlds))
)
appconf.meta.append(FieldConf(id="testfld", datatype="number", size=4, name="Number"))
appconf.meta.append(FieldConf(id="title", datatype="string", size=100, name="Title"))
appconf.meta.append(FieldConf(id="pool_sort", datatype="number", size=4, name="Number"))
appconf.meta.append(FieldConf(id="pool_wfa", datatype="string", size=20, name="Workflow"))
appconf.meta.append(FieldConf(id="pool_wfp", datatype="string", size=20, name="Workflow"))

# test data -----------------------------------------------------------------------
data1_1 = { "ftext": "this is text!",
            "fnumber": 123456,
            "fdate": "2008/06/23 16:55:00",
            "ftime": "16:55:00",
            "flist": ["item 1", "item 2", "item 3"],
            "fmselect": "item 5",
            "funit": 35,
            "funitlist": [34, 35, 36],
            "pool_filename":"äüöß and others",
            "pool_type": "type1"}
data2_1 = { "fstr": "this is sting!",
            "ftext": "this is text!",
            "pool_filename": "äüöß and others",
            "pool_type": "type2"}
data3_1 = { "pool_filename": "title data 3",
            "fstr": "testing type 3!"}

data1_2 = { "ftext": "this is a new text!",
            "ftime": "23:08:13.000500",
            "funit": 0,
            "pool_filename":"new title data 1"}
data2_2 = { "fstr": "this is new sting!",
            "pool_filename": "new title data 2"}

file2_1_data=b"This is the first text"
file2_2_data=b"This is the text in the second file"
file2_1 = {"filename":"file1.txt", "file":file2_1_data}
file2_2 = {"filename":"file2.txt", "file":file2_2_data}


# empty -------------------------------------------------------------------------
def app_db(modules=None):
    app = Application()
    app.Register(appconf)
    if modules:
        for m in modules:
            app.Register(m)
    p = Portal()
    p.Register(app, "nive")
    app.SetupApplication()
    dbfile = DvPath(app.dbConfiguration.dbName)
    if not dbfile.IsFile():
        dbfile.CreateDirectories()
    root = DvPath(app.dbConfiguration.fileRoot)
    if not root.IsDirectory():
        root.CreateDirectories()

    """
    db = app.db
    try:
        cursor = db.Execute("select id from pool_meta where id=1")
        db.Execute("select id from data1 where id=1", cursor=cursor)
        db.Execute("select id from data2 where id=1", cursor=cursor)
        db.Execute("select id from data3 where id=1", cursor=cursor)
        db.Execute("select id from pool_files where id=1", cursor=cursor)
        db.Execute("select id from pool_sys where id=1", cursor=cursor)
        db.Execute("select id from pool_groups where id=1", cursor=cursor)
        db.Execute("select title from pool_meta where id=1", cursor=cursor)
        cursor.close()
    except:
        app.GetTool("nive.tools.dbStructureUpdater")()
    """

    # disable this to update test tables each time the tests are called
    app.GetTool("nive.tools.dbStructureUpdater")()
    app.Startup(None)
    return app

def app_nodb():
    app = Application()
    app.Register(appconf)
    app.Register(DatabaseConf())
    p = Portal()
    p.Register(app, "nive")
    app.SetupApplication()
    try:
        app.Startup(None)
    except OperationalError:
        pass
    return app

def emptypool(app, files=False):
    db = app.db
    cursor = db.Execute("delete FROM pool_meta")
    db.Execute("delete FROM pool_files", cursor=cursor)
    db.Execute("delete FROM pool_fulltext", cursor=cursor)
    db.Execute("delete FROM pool_groups", cursor=cursor)
    db.Execute("delete FROM pool_sys", cursor=cursor)
    db.Execute("delete FROM data3", cursor=cursor)
    db.Execute("delete FROM data2", cursor=cursor)
    db.Execute("delete FROM data1", cursor=cursor)
    db.Commit()
    if files:
        import shutil
        shutil.rmtree(str(db.root), ignore_errors=True)
        db.root.CreateDirectories()

def createpool(path, app):
    path.CreateDirectories()
    app.GetTool("nive.tools.dbStructureUpdater")()


def statdb(app):
    c = app.db.GetCountEntries()
    ##print "Count entries in DB:", c
    return c

def root(a):
    r = a.GetRoot("root")
    return r

def createObj1(c):
    type = "type1"
    data = data1_1
    user = User("test")
    o = c.Create(type, data = data, user = user)
    return o

def createObj2(c):
    type = "type2"
    data = data2_1
    user = User("test")
    o = c.Create(type, data = data, user = user)
    return o

def createObj3(c):
    type = "type3"
    data = data3_1
    user = User("test")
    o = c.Create(type, data = data, user = user)
    return o

def createObj1file(c):
    type = "type1"
    data = data1_1.copy()
    data["file1"] = File(**file2_1)
    data["file2"] = File(**file2_2)
    user = User("test")
    o = c.Create(type, data = data, user = user)
    return o

def createObj2file(c):
    type = "type2"
    data = data2_1.copy()
    data["file2"] = File(**file2_1)
    user = User("test")
    o = c.Create(type, data = data, user = user)
    return o

def maxobj(r):
    id = r.GetMaxID()
    return r.obj(id)


def populate(testobj):
    # create three levels of entries
    level1 = 5
    level2 = 5
    level3_1 = 5
    level3_2 = 5
    c = testobj.app.root
    ids = []
    n = 0
    for i in range(0, level1):
        o = createObj1(c)
        n += 1
        ids.append(o.id)
        for i2 in range(0, level2):
            o2 = createObj1(o)
            n += 1
            for i3 in range(0, level3_1):
                createObj1(o2)
                n += 1
            for i3 in range(0, level3_2):
                o4 = createObj2(o2)
                id = o4.id
                n += 1
    testobj.ids = ids
    testobj.lastid = id

